from ..pair_counters import SimulationBoxPairCount, SurveyDataPairCount
from .estimators import LandySzalayEstimator, NaturalEstimator, WedgeBinnedStatistic
from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic

import numpy
import warnings
import logging

class BasePairCount2PCF(object):
    """
    Base class for two-point correlation function algorithms that use
    pair counting. The API largely follows that of
    :class:`~nbodykit.algorithms.SimulationBoxPairCount` and
    :class:`~nbodykit.algorithms.SurveyDataPairCount`.

    Parameters
    ----------
    mode : '1d', '2d', 'projected', 'angular'
        the type of two point correlation function to compute
    data1 : CatalogSource
        the data catalog; must have a 'Position' column
    edges : array_like
        the bin edges along the first binning dimension
    Nmu : int, optional
        when ``mode`` is '2d', the number of mu bins, ranging from 0 to 1
    pimax : float, optional
        when ``mode`` is 'projected', the maximum separation along the line-of-sight
    randoms1 : CatalogSource, optional
        the catalog specifying the un-clustered, random distribution for ``data1``;
        if not provided, analytic randoms will be used
    randoms2 : CatalogSource, optional
        the catalog specifying the un-clustered, random distribution for ``data2``;
        if not provided, analytic randoms will be used
    data2 : CatalogSource, optional
        the second data catalog to cross-correlate; must have a 'Position' column
    R1R2 : SimulationBoxPairCount, SurveyDataPairCount, optional
        if provided, random pairs R1R2 are not recalculated in the Landy-Szalay estimator
    **kws :
        additional keyword arguments passed to the appropriate pair counting class
    """

    def __init__(self, mode, data1, edges,
                    Nmu=None, pimax=None,
                    randoms1=None, randoms2=None, data2=None, R1R2=None, **kws):

        self.comm = data1.comm

        # store the attributes
        self.attrs = {'mode':mode, 'edges':edges, 'Nmu':Nmu, 'pimax':pimax}
        self.attrs.update(kws)

        # store the catalogs
        self.data1 = data1
        self.data2 = data2
        self.randoms1 = randoms1
        self.randoms2 = randoms2
        self.R1R2 = R1R2


    def run(self):
        """
        Run the two-point correlation function algorithm.

        There are two cases here:

        1. If no randoms were provided, and the data is in a simulation box with
           periodic boundary conditions, the natural estimator
           :math:`DD/RR - 1` is used.
        2. If randoms were provided, the Landy-Szalay estimator is used:
           :math:`(D_1 D_2 - D_1 R_2 - D_2 R_1 + R_1 R_2) / R_1 R_2`

        Raises
        ------
        ValueError :
            if periodic boundary conditions were not requested, and ``randoms1``
            is ``None``
        ValueError :
            if periodic boundary conditions were not requested, and
            ``data2`` is not None, but ``randoms2`` is ``None``
        """
        # get the config
        attrs = self.attrs.copy()
        config = attrs.pop('config')
        attrs.update(config)

        # whether we are doing sim volume or mock survey
        if 'periodic' in attrs:
            pair_counter = SimulationBoxPairCount
        else:
            pair_counter = SurveyDataPairCount

        # use analytic randoms for a periodic box
        if 'periodic' in attrs and attrs['periodic'] and self.randoms1 is None:

            if attrs['mode'] == 'angular':
                self.data1 = _restrict_to_spherical_volume(self.data1)
                if self.data2 is not None:
                    self.data2 = _restrict_to_spherical_volume(self.data2)
                if self.comm.rank == 0:
                    msg = "when using analytic randoms for the angular correlation function, "
                    msg += "we restrict the input data to a spherical volume, throwing away objects. "
                    msg += "To use all data, pass in a UniformCatalog as the 'randoms1' keyword"
                    warnings.warn(msg)

            # count the data-data pairs using analytic randoms
            DD = pair_counter(first=self.data1, second=self.data2, **attrs)
            self.R1R2, self.corr = NaturalEstimator(DD)
            self.D1D2 = DD.pairs
            self.D1R2 = self.D2R1 = None

        # need catalog randoms
        else:

            if self.randoms1 is None:
                msg = "a catalog of randoms must be specified as the ``randoms1`` keyword "
                msg += "when the data is not in a simulation box with periodic boundary conditions"
                raise ValueError(msg)

            # use the first randoms for both
            if self.data2 is not None and self.randoms2 is None:
                self.randoms2 = self.randoms1

            # use the Landy-Szalay estimator
            result = LandySzalayEstimator(pair_counter, self.data1, self.data2,
                                            self.randoms1, self.randoms2, R1R2=self.R1R2,
                                            logger=self.logger, **attrs)
            self.D1D2, self.D1R2, self.D2R1, self.R1R2, self.corr = result

    def __getstate__(self):

        # the correlation
        edges = [self.corr.edges[d] for d in self.corr.dims]
        state = {'corr':self.corr.data, 'dims':self.corr.dims, 'edges':edges}

        # the pair counts
        for pc in ['D1D2', 'D1R2', 'D2R1', 'R1R2', 'wp']:
            if getattr(self, pc, None) is not None:
                state[pc] = getattr(self, pc).data
            else:
                state[pc] = None

        state['attrs'] = self.attrs
        return state

    def __setstate__(self, state):

        edges = state.pop('edges')
        dims = state.pop('dims')
        self.__dict__.update(state)

        self.corr = WedgeBinnedStatistic(dims, edges, self.corr)
        if self.wp is not None:
            # NOTE: only edges[0], second dimension was summed over
            self.wp = WedgeBinnedStatistic(dims[:1], edges[:1], self.wp)

        for pc in ['D1D2', 'D1R2', 'D2R1', 'R1R2']:
            val = getattr(self, pc)
            if val is not None:
                setattr(self, pc, WedgeBinnedStatistic(dims, edges, val))

    def save(self, output):
        """
        Save result as a JSON file with name ``output``
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('measurement done; saving result to %s' % output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, output, comm=None):
        """
        Load a result has been saved to disk with :func:`save`.
        """
        import json
        from nbodykit.utils import JSONDecoder
        if comm.rank == 0:
            with open(output, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)
        self.__setstate__(state)
        self.comm = comm
        return self


class SimulationBox2PCF(BasePairCount2PCF):
    r"""
    Compute the two-point correlation function for data in a simulation box
    as a function of :math:`r`, :math:`(r,\mu)`, :math:`(r_p, \pi)`, or
    :math:`\theta` using pair counting.

    This uses analytic randoms when using periodic conditions, unless
    a randoms catalog is specified. The "natural" estimator (DD/RR-1) is
    used in the former case, and the Landy-Szalay estimator (DD/RR - 2DR/RR + 1)
    in the latter case.

    .. note::
        When using analytic randoms, the expected counts are assumed to
        be unweighted.

    Parameters
    ----------
    mode : '1d', '2d', 'projected', 'angular'
        the type of two-point correlation function to compute; see the Notes below
    data1 : CatalogSource
        the data catalog; must have a 'Position' column
    edges : array_like
        the separation bin edges along the first coordinate dimension;
        depending on ``mode``, the options are :math:`r`, :math:`r_p`, or
        :math:`\theta`. Expected units for distances are :math:`\mathrm{Mpc}/h`
        and degrees for angles. Length of nbins+1
    Nmu : int, optional
        the number of :math:`\mu` bins, ranging from 0 to 1; requred if
        ``mode='2d'``
    pimax : float, optional
        The maximum separation along the line-of-sight when ``mode='projected'``.
        Distances along the :math:`\pi` direction are binned with unit
        depth. For instance, if ``pimax=40``, then 40 bins will be created
        along the :math:`\pi` direction.
    data2 : CatalogSource, optional
        the second data catalog to cross-correlate; must have a 'Position' column
    randoms1 : CatalogSource, optional
        the catalog specifying the un-clustered, random distribution for ``data1``;
        if not provided, analytic randoms will be used
    randoms2 : CatalogSource, optional
        the catalog specifying the un-clustered, random distribution for ``data2``;
        if not provided, analytic randoms will be used
    R1R2 : SimulationBoxPairCount, optional
        if provided, random pairs R1R2 are not recalculated in the Landy-Szalay estimator
    periodic : bool, optional
        whether to use periodic boundary conditions
    BoxSize : float, 3-vector, optional
        the size of the box; if 'BoxSize' is not provided in the source
        'attrs', it must be provided here
    los : 'x', 'y', 'z'; int, optional
        the axis of the simulation box to treat as the line-of-sight direction;
        this can be provided as string identifying one of 'x', 'y', 'z' or
        the equivalent integer number of the axis
    weight : str, optional
        the name of the column in the source specifying the particle weights
    show_progress : bool, optional
        if ``True``, perform the pair counting calculation in 10 iterations,
        logging the progress after each iteration; this is useful for
        understanding the scaling of the code
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function

    Notes
    -----
    This class can compute correlation functions using several different
    coordinate choices, based on the value of the input argument ``mode``.
    The choices are:

    * ``mode='1d'`` : compute pairs as a function of the 3D separation :math:`r`
    * ``mode='2d'`` : compute pairs as a function of the 3D separation :math:`r`
      and the cosine of the angle to the line-of-sight, :math:`\mu`
    * ``mode='projected'`` : compute pairs as a function of distance perpendicular
      and parallel to the line-of-sight, :math:`r_p` and :math:`\pi`
    * ``mode='angular'`` : compute pairs as a function of angle on the sky, :math:`\theta`

    If ``mode='projected'``, the projected correlation function :math:`w_p(r_p)`
    is also computed, using the input :math:`\pi_\mathrm{max}` value.
    """
    logger = logging.getLogger('SimulationBox2PCF')

    def __init__(self, mode, data1, edges, Nmu=None, pimax=None,
                    data2=None, randoms1=None, randoms2=None, R1R2=None,
                    periodic=True, BoxSize=None, los='z',
                    weight='Weight', show_progress=False, **config):

        # format the input arguments
        args = dict(locals())
        args.pop('self')

        # init the base class
        BasePairCount2PCF.__init__(self, **args)

        # and run
        self.run()

    def run(self):
        """
        Run the two-point correlation function algorithm. This attaches
        the following attributes:

        - :attr:`SimulationBox2PCF.D1D2`
        - :attr:`SimulationBox2PCF.D1R2`
        - :attr:`SimulationBox2PCF.D2R1`
        - :attr:`SimulationBox2PCF.R1R2`
        - :attr:`SimulationBox2PCF.corr`
        - :attr:`SimulationBox2PCF.wp` (if ``mode='projected'``)

        Attributes
        ----------
        D1D2 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the data1 - data2 pair counts
        D1R2 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the data1 - randoms2 pair counts
        D2R1 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the data2 - randoms1 pair counts
        R1R2 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the randoms1 - randoms2 pair counts
        corr : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the correlation function values, stored as the ``corr`` variable,
            computed from the pair counts
        wp : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the projected correlation function, :math:`w_p(r_p)`, computed
            if ``mode='projected'``; correlation is stored as the ``corr`` variable

        Notes
        -----
        The :attr:`D1D2`, :attr:`D1R2`, :attr:`D2R1`, and :attr:`R1R2`
        attributes are identical to the
        :attr:`~nbodykit.algorithms.SimulationBoxPairCount.pairs` attribute
        of :class:`~nbodykit.algorithms.SimulationBoxPairCount`.
        """
        # this does most of the work
        BasePairCount2PCF.run(self)

        # compute wp(rp) if we computed xi(rp, pi)
        if self.attrs['mode'] == 'projected':
            self.wp = _compute_wp(self.corr)


class SurveyData2PCF(BasePairCount2PCF):
    r"""
    Compute the two-point correlation function for observational survey data
    as a function of :math:`r`, :math:`(r,\mu)`, :math:`(r_p, \pi)`, or
    :math:`\theta` using pair counting.

    The Landy-Szalay estimator (DD/RR - 2 DD/RR + 1) is used to transform
    pair counts in to the correlation function.

    Parameters
    ----------
    mode : '1d', '2d', 'projected', 'angular'
        the type of two-point correlation function to compute; see the Notes below
    data1 : CatalogSource
        the data catalog; must have a 'Position' column
    randoms1 : CatalogSource
        the catalog specifying the un-clustered, random distribution for ``data1``
    edges : array_like
        the separation bin edges along the first coordinate dimension;
        depending on ``mode``, the options are :math:`r`, :math:`r_p`, or
        :math:`\theta`. Expected units for distances are :math:`\mathrm{Mpc}/h`
        and degrees for angles. Length of nbins+1
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`, optional
        the cosmology instance used to convert redshift into comoving distance;
        this is required for all cases except ``mode='angular'``
    Nmu : int, optional
        the number of :math:`\mu` bins, ranging from 0 to 1; requred if
        ``mode='2d'``
    pimax : float, optional
        The maximum separation along the line-of-sight when ``mode='projected'``.
        Distances along the :math:`\pi` direction are binned with unit
        depth. For instance, if ``pimax=40``, then 40 bins will be created
        along the :math:`\pi` direction.
    data2 : CatalogSource, optional
        the second data catalog to cross-correlate; must have a 'Position' column
    randoms2 : CatalogSource, optional
        the catalog specifying the un-clustered, random distribution for ``data2``;
        if not specified and ``data2`` is provied, then ``randoms1`` will be used
        for both.
    R1R2 : SurveyDataPairCount, optional
        if provided, random pairs R1R2 are not recalculated in the Landy-Szalay estimator
    ra : str, optional
        the name of the column in the source specifying the
        right ascension coordinates in units of degrees; default is 'RA'
    dec : str, optional
        the name of the column in the source specifying the declination
        coordinates; default is 'DEC'
    redshift : str, optional
        the name of the column in the source specifying the redshift
        coordinates; default is 'Redshift'
    weight : str, optional
        the name of the column in the source specifying the object weights
    show_progress : bool, optional
        if ``True``, perform the pair counting calculation in 10 iterations,
        logging the progress after each iteration; this is useful for
        understanding the scaling of the code
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function

    Notes
    -----
    This class can compute correlation functions using several different
    coordinate choices, based on the value of the input argument ``mode``.
    The choices are:

    * ``mode='1d'`` : compute pairs as a function of the 3D separation :math:`r`
    * ``mode='2d'`` : compute pairs as a function of the 3D separation :math:`r`
      and the cosine of the angle to the line-of-sight, :math:`\mu`
    * ``mode='projected'`` : compute pairs as a function of distance perpendicular
      and parallel to the line-of-sight, :math:`r_p` and :math:`\pi`
    * ``mode='angular'`` : compute pairs as a function of angle on the sky, :math:`\theta`

    If ``mode='projected'``, the projected correlation function :math:`w_p(r_p)`
    is also computed, using the input :math:`\pi_\mathrm{max}` value.
    """
    logger = logging.getLogger('SurveyData2PCF')

    def __init__(self, mode, data1, randoms1, edges, cosmo=None,
                    Nmu=None, pimax=None, data2=None, randoms2=None, R1R2=None,
                    ra='RA', dec='DEC', redshift='Redshift', weight='Weight',
                    show_progress=False, **config):

        # format the input arguments
        args = dict(locals())
        args.pop('self')

        # init the base class
        BasePairCount2PCF.__init__(self, **args)

        # and run
        self.run()

    def run(self):
        """
        Run the two-point correlation function algorithm. This attaches
        the following attributes:

        - :attr:`SurveyData2PCF.D1D2`
        - :attr:`SurveyData2PCF.D1R2`
        - :attr:`SurveyData2PCF.D2R1`
        - :attr:`SurveyData2PCF.R1R2`
        - :attr:`SurveyData2PCF.corr`
        - :attr:`SurveyData2PCF.wp` (if ``mode='projected'``)

        Attributes
        ----------
        D1D2 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the data1 - data2 pair counts
        D1R2 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the data1 - randoms2 pair counts
        D2R1 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the data2 - randoms1 pair counts
        R1R2 : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the randoms1 - randoms2 pair counts
        corr : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the correlation function values, stored as the ``corr`` variable,
            computed from the pair counts
        wp : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            the projected correlation function, :math:`w_p(r_p)`, computed
            if ``mode='projected'``; correlation is stored as the ``corr`` variable

        Notes
        -----
        The :attr:`D1D2`, :attr:`D1R2`, :attr:`D2R1`, and :attr:`R1R2`
        attributes are identical to the
        :attr:`~nbodykit.algorithms.SurveyDataPairCount.pairs` attribute
        of :class:`~nbodykit.algorithms.SurveyDataPairCount`.
        """
        # this does most of the work
        BasePairCount2PCF.run(self)

        # compute wp(rp) if we computed xi(rp, pi)
        if self.attrs['mode'] == 'projected':
            self.wp = _compute_wp(self.corr)


def _compute_wp(corr):
    r"""
    Compute the projected correlation function :math:`w_p(r_p)` from
    :math:`\xi(r_p, \pi)`.
    """
    # compute wp(rp)
    dpi = numpy.diff(corr.edges['pi'])
    wp = (2*corr['corr']*dpi).sum(axis=-1)

    # return a BinnedStatistic
    toret = corr.copy()
    if len(toret.dims) > 1:
        toret = toret.average(toret.dims[-1])
    toret['corr'] = wp

    return toret

def _restrict_to_spherical_volume(source):
    """
    Slice the input source such that it only includes a spherical volume
    instead the simulation box.

    This returns a CatalogSource only containing objects within a radius
    of 1/2 the minimum box side length of the box center.
    """
    # compute the distance from box center
    origin = 0.5 * source.attrs['BoxSize']
    pos = source['Position'] - origin
    r = (pos**2).sum(axis=-1)**0.5

    # restrict to sphere less than half of box size
    keep = r < 0.5 * source.attrs['BoxSize'].min()
    return source[keep]



from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic
from collections import OrderedDict
import numpy
import logging
import kdcount

class Base3PCF(object):
    """
    Base class for implementing common 3PCF calculations.

    Users should use :class:`SimulationBox3PCF` or :class:`SurveyData3PCF`.
    """
    def __init__(self, source, poles, edges, required_cols, BoxSize=None, periodic=None):

        from .pair_counters.base import verify_input_sources

        # verify the input sources
        inspect = periodic is not None
        BoxSize = verify_input_sources(source, None, BoxSize, required_cols, inspect_boxsize=inspect)

        self.source = source
        self.comm = self.source.comm

        # save the meta-data
        self.attrs = {}
        self.attrs['poles'] = poles
        self.attrs['edges'] = edges

        # store periodic/BoxSize for SimulationBox
        if periodic is not None:
            self.attrs['BoxSize'] = BoxSize
            self.attrs['periodic'] = periodic

    def _run(self, pos, w, pos_sec, w_sec, boxsize=None, bunchsize=10000):
        """
        Internal function to run the 3PCF algorithm on the input data and
        weights.

        The input data/weights have already been domain-decomposed, and
        the loads should be balanced on all ranks.
        """
        # maximum radius
        rmax = numpy.max(self.attrs['edges'])

        # the array to hold output values
        nbins  = len(self.attrs['edges'])-1
        Nell   = len(self.attrs['poles'])
        zeta = numpy.zeros((Nell,nbins,nbins), dtype='f8')
        alms = {}
        walms = {}

        # compute the Ylm expressions we need
        if self.comm.rank == 0:
            self.logger.info("computing Ylm expressions...")
        Ylm_cache = YlmCache(self.attrs['poles'], self.comm)
        if self.comm.rank ==  0:
            self.logger.info("...done")

        # make the KD-tree holding the secondaries
        tree_sec = kdcount.KDTree(pos_sec, boxsize=boxsize).root

        def callback(r, i, j, iprim=None):

            # remove self pairs
            valid = r > 0.
            r = r[valid]; i = i[valid]

            # normalized, re-centered position array (periodic)
            dpos = (pos_sec[i] - pos[iprim])

            # enforce periodicity in dpos
            if boxsize is not None:
                for axis, col in enumerate(dpos.T):
                    col[col > boxsize[axis]*0.5] -= boxsize[axis]
                    col[col <= -boxsize[axis]*0.5] += boxsize[axis]
            recen_pos = dpos / r[:,numpy.newaxis]

            # find the mapping of r to rbins
            dig = numpy.searchsorted(self.attrs['edges'], r, side='left')

            # evaluate all Ylms
            Ylms = Ylm_cache(recen_pos[:,0]+1j*recen_pos[:,1], recen_pos[:,2])

            # sqrt of primary weight
            w0 = w[iprim]

            # loop over each (l,m) pair
            for (l,m) in Ylms:

                # the Ylm evaluated at galaxy positions
                weights = Ylms[(l,m)] * w_sec[i]

                # sum over for each radial bin
                alm = alms.setdefault((l, m), numpy.zeros(nbins, dtype='c16'))
                walm = walms.setdefault((l, m), numpy.zeros(nbins, dtype='c16'))

                r1 = numpy.bincount(dig, weights=weights.real, minlength=nbins+2)[1:-1]
                alm[...] += r1
                walm[...] += w0 * r1
                if m != 0:
                    i1 = numpy.bincount(dig, weights=weights.imag, minlength=nbins+2)[1:-1]
                    alm[...] += 1j*i1
                    walm[...] += w0*1j*i1

        # determine rank with largest load
        loads = self.comm.allgather(len(pos))
        largest_load = numpy.argmax(loads)
        chunk_size = max(loads) // 10

        # compute multipoles for each primary (s vector in the paper)
        for iprim in range(len(pos)):
            # alms must be clean for each primary particle; (s) in eq 15 and 8 of arXiv:1506.02040v2
            alms.clear()
            walms.clear()
            tree_prim = kdcount.KDTree(numpy.atleast_2d(pos[iprim]), boxsize=boxsize).root
            tree_sec.enum(tree_prim, rmax, process=callback, iprim=iprim, bunch=bunchsize)

            if self.comm.rank == largest_load and iprim % chunk_size == 0:
                self.logger.info("%d%% done" % (10*iprim//chunk_size))

            # combine alms into zeta(s);
            # this cannot be done in the callback because
            # it is a nonlinear function (outer product) of alm.
            for (l, m) in alms:
                alm = alms[(l, m)]
                walm = walms[(l, m)]

                # compute alm * conjugate(alm)
                alm_w_alm = numpy.outer(walm, alm.conj())
                if m != 0: alm_w_alm += alm_w_alm.T # add in the -m contribution for m != 0
                zeta[Ylm_cache.ell_to_iell[l], ...] += alm_w_alm.real

        # sum across all ranks
        zeta = self.comm.allreduce(zeta)

        # normalize according to Eq. 15 of Slepian et al. 2015
        # differs by factor of (4 pi)^2 / (2l+1) from the C++ code
        zeta /= (4*numpy.pi)

        # make a BinnedStatistic
        dtype = numpy.dtype([('corr_%d' % ell, zeta.dtype) for ell in self.attrs['poles']])
        data = numpy.empty(zeta.shape[-2:], dtype=dtype)
        for i, ell in enumerate(self.attrs['poles']):
            data['corr_%d' % ell] = zeta[i]

        # save the result
        edges = self.attrs['edges']
        poles = BinnedStatistic(['r1', 'r2'], [edges, edges], data)
        return poles

    def __getstate__(self):
        return {'poles':self.poles.data, 'attrs':self.attrs}

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.poles = BinnedStatistic(['r1', 'r2'], [self.attrs['edges']]*2, self.poles)

    def save(self, output):
        """
        Save the :attr:`poles` result to a JSON file with name ``output``.
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('measurement done; saving result to %s' % output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, filename, comm=None):
        """
        Load a result from ``filename`` that has been saved to
        disk with :func:`save`.
        """
        import json
        from nbodykit.utils import JSONDecoder
        if comm.rank == 0:
            with open(filename, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)
        self.__setstate__(state)
        self.comm = comm
        return self

class SimulationBox3PCF(Base3PCF):
    """
    Compute the multipoles of the isotropic, three-point correlation function
    in configuration space for data in a simulation box.

    This uses the algorithm of Slepian and Eisenstein, 2015 which scales
    as :math:`\mathcal{O}(N^2)`, where :math:`N` is the number of objects.

    Results are computed when the object is inititalized. See the documenation
    of :func:`run` for the attributes storing the results.

    .. note::

        The algorithm expects the positions of objects in a simulation box to
        be the Cartesian ``x``, ``y``, and ``z`` vectors. For survey data,
        in the form of right ascension, declination, and
        redshift, see :class:`~nbodykit.algorithms.SurveyData3PCF`.

    Parameters
    ----------
    source : CatalogSource
        the input source of particles providing the 'Position' column
    poles : list of int
        the list of multipole numbers to compute
    edges : array_like
        the edges of the bins of separation to use; length of nbins+1
    BoxSize : float, 3-vector, optional
        the size of the box; if periodic boundary conditions used, and 'BoxSize'
        not provided in the source :attr:`attrs`, it must be provided here
    periodic : bool, optional
        whether to use periodic boundary conditions when computing separations
        between objects
    weight : str, optional
        the name of the column in the source specifying the particle weights

    References
    ----------
    Slepian and Eisenstein, MNRAS 454, 4142-4158 (2015)
    """
    logger = logging.getLogger("SimulationBox3PCF")

    def __init__(self, source, poles, edges, BoxSize=None, periodic=True, weight='Weight'):

        # initialize the base class
        required_cols = ['Position', weight]
        Base3PCF.__init__(self, source, poles, edges, required_cols,
                            BoxSize=BoxSize, periodic=periodic)

        # save the weight column
        self.attrs['weight'] = weight

        # check largest possible separation
        if periodic:
            min_box_side = 0.5*self.attrs['BoxSize'].min()
            if numpy.amax(edges) > min_box_side:
                raise ValueError(("periodic pair counts cannot be computed for Rmax > BoxSize/2"))

        # run the algorithm
        self.poles = self.run()


    def run(self, pedantic=False):
        """
        Compute the three-point CF multipoles. This attaches the following
        the attributes to the class:

        - :attr:`poles`

        Attributes
        ----------
        poles : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object to hold the multipole results; the
            binned statistics stores the multipoles as variables ``corr_0``,
            ``corr_1``, etc for :math:`\ell=0,1,` etc. The coordinates
            of the binned statistic are ``r1`` and ``r2``, which give the
            separations between the three objects in CF.
        """
        from .pair_counters.domain import decompose_box_data

        # the box size to use
        if self.attrs['periodic']:
            boxsize = self.attrs['BoxSize']
        else:
            boxsize = None

        # domain decompose the data
        smoothing = numpy.max(self.attrs['edges'])
        (pos, w), (pos_sec, w_sec) = decompose_box_data(self.source, None, self.attrs,
                                                        self.logger, smoothing)

        # run the algorithm
        if pedantic:
            return self._run(pos, w, pos_sec, w_sec, boxsize=boxsize, bunchsize=1)
        else:
            return self._run(pos, w, pos_sec, w_sec, boxsize=boxsize)

class SurveyData3PCF(Base3PCF):
    """
    Compute the multipoles of the isotropic, three-point correlation function
    in configuration space for observational survey data.

    This uses the algorithm of Slepian and Eisenstein, 2015 which scales
    as :math:`\mathcal{O}(N^2)`, where :math:`N` is the number of objects.

    Results are computed when the object is inititalized. See the documenation
    of :func:`run` for the attributes storing the results.

    .. note::

        The algorithm expects the positions of objects from a survey catalog
        be the sky coordinates, right ascension and declination, and redshift.
        For simulation box data in Cartesian coordinates, see
        :class:`~nbodykit.algorithms.SimulationBox3PCF`.

    .. warning::
        The right ascension and declination columns should be specified
        in degrees.

    Parameters
    ----------
    source : CatalogSource
        the input source of particles providing the 'Position' column
    poles : list of int
        the list of multipole numbers to compute
    edges : array_like
        the edges of the bins of separation to use; length of nbins+1
    cosmo : :class:`~nbodykit.cosmology.cosmology.Cosmology`
        the cosmology instance used to convert redshifts into comoving distances
    ra : str, optional
        the name of the column in the source specifying the
        right ascension coordinates in units of degrees; default is 'RA'
    dec : str, optional
        the name of the column in the source specifying the declination
        coordinates; default is 'DEC'
    redshift : str, optional
        the name of the column in the source specifying the redshift
        coordinates; default is 'Redshift'
    weight : str, optional
        the name of the column in the source specifying the object weights
    domain_factor : int, optional
        the integer value by which to oversubscribe the domain decomposition
        mesh before balancing loads; this number can affect the distribution
        of loads on the ranks -- an optimal value will lead to balanced loads

    References
    ----------
    Slepian and Eisenstein, MNRAS 454, 4142-4158 (2015)
    """
    logger = logging.getLogger("SurveyData3PCF")

    def __init__(self, source, poles, edges, cosmo, domain_factor=4,
                    ra='RA', dec='DEC', redshift='Redshift', weight='Weight'):

        # initialize the base class
        required_cols = [ra, dec, redshift, weight]
        Base3PCF.__init__(self, source, poles, edges, required_cols)

        # save meta-data
        self.attrs['cosmo'] = cosmo
        self.attrs['weight'] = weight
        self.attrs['ra'] = ra
        self.attrs['dec'] = dec
        self.attrs['redshift'] = redshift
        self.attrs['domain_factor'] = domain_factor

        # run the algorithm
        self.poles = self.run()

    def run(self):
        """
        Compute the three-point CF multipoles. This attaches the following
        the attributes to the class:

        - :attr:`poles`

        Attributes
        ----------
        poles : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object to hold the multipole results; the
            binned statistics stores the multipoles as variables ``corr_0``,
            ``corr_1``, etc for :math:`\ell=0,1,` etc. The coordinates
            of the binned statistic are ``r1`` and ``r2``, which give the
            separations between the three objects in CF.
        """
        from .pair_counters.domain import decompose_survey_data

        # domain decompose the data
        # NOTE: pos and pos_sec are Cartesian!
        smoothing = numpy.max(self.attrs['edges'])
        (pos, w), (pos_sec, w_sec) = decompose_survey_data(self.source, None,
                                                            self.attrs, self.logger,
                                                            smoothing,
                                                            return_cartesian=True,
                                                            domain_factor=self.attrs['domain_factor'])

        # run the algorithm
        return self._run(pos, w, pos_sec, w_sec)


class YlmCache(object):
    """
    A class to compute spherical harmonics :math:`Y_{lm}` up
    to a specified maximum :math:`\ell`.

    During calculation, the necessary power of Cartesian unit
    vectors are cached in memory to avoid repeated calculations
    for separate harmonics.
    """
    def __init__(self, ells, comm):

        import sympy as sp
        from sympy.utilities.lambdify import implemented_function
        from sympy.parsing.sympy_parser import parse_expr

        self.ells = numpy.asarray(ells).astype(int)
        self.max_ell = max(self.ells)

        # look up table from ell to iell, index for cummulating results.
        self.ell_to_iell = numpy.empty(self.max_ell + 1, dtype=int)
        for iell, ell in enumerate(self.ells):
            self.ell_to_iell[ell] = iell

        lms = [(l,m) for l in ells for m in range(0, l+1)]

        # compute the Ylm string expressions in parallel
        exprs = []
        for i in range(comm.rank, len(lms), comm.size):
            lm = lms[i]
            exprs.append((lm, str(self._get_Ylm(*lm))))
        exprs = [x for sublist in comm.allgather(exprs) for x in sublist]

        # determine the powers entering into each expression
        args = {}
        for lm, expr in exprs:
            matches = []
            for var in ['xpyhat', 'zhat']:
                for e in range(2, max(ells)+1):
                    name = var + '**' + str(e)
                    if name in expr:
                        matches.append((name, 'cached_'+var, str(e)))
                args[lm] = matches


        # define a function to return cached power
        def from_cache(name, pow):
            return self._cache[str(name)+str(pow)]
        f = implemented_function(sp.Function('from_cache'), from_cache)

        # arguments to the sympy functions
        zhat   = sp.Symbol('zhat', real=True, positive=True)
        xpyhat = sp.Symbol('xpyhat', complex=True)

        self._cache = {}

        # make the Ylm functions
        self._Ylms = OrderedDict()
        for lm, expr in exprs:
            expr = parse_expr(expr, local_dict={'zhat':zhat, 'xpyhat':xpyhat})
            for var in args[lm]:
                expr = expr.replace(var[0], 'from_cache(%s, %s)' %var[1:])
            self._Ylms[lm] = sp.lambdify((xpyhat, zhat), expr)

    def __call__(self, xpyhat, zhat):
        """
        Return a dictionary holding Ylm for each (l,m) combination
        required

        Parameters
        ----------
        xpyhat : array_like
            a complex array holding xhat + i * yhat, where xhat and yhat
            are the two cartesian unit vectors
        zhat : array_like
            the third cartesian unit vector
        """
        # fill the cache first
        self._cache['cached_xpyhat2'] = xpyhat**2
        self._cache['cached_zhat2'] = zhat**2
        for name,x in zip(['cached_xpyhat', 'cached_zhat'], [xpyhat, zhat]):
            for i in range(3, self.max_ell+1):
                self._cache[name+str(i)] = self._cache[name+str(i-1)]*x

        # return a dictionary for each (l,m) tuple
        toret = {}
        for lm in self._Ylms:
            toret[lm] = self._Ylms[lm](xpyhat, zhat)
        return toret

    def _get_Ylm(self, l, m):
        """
        Compute an expression for spherical harmonic of order (l,m)
        in terms of Cartesian unit vectors, :math:`\hat{z}`
        and :math:`\hat{x} + i \hat{y}`

        Parameters
        ----------
        l : int
            the degree of the harmonic
        m : int
            the order of the harmonic; |m| < l

        Returns
        -------
        expr :
            a sympy expression that corresponds to the
            requested Ylm

        References
        ----------
        https://en.wikipedia.org/wiki/Spherical_harmonics
        """
        import sympy as sp

        # the relevant cartesian and spherical symbols
        x, y, z, r = sp.symbols('x y z r', real=True, positive=True)
        xhat, yhat, zhat = sp.symbols('xhat yhat zhat', real=True, positive=True)
        xpyhat = sp.Symbol('xpyhat', complex=True)
        phi, theta = sp.symbols('phi theta')
        defs = [(sp.sin(phi), y/sp.sqrt(x**2+y**2)),
                (sp.cos(phi), x/sp.sqrt(x**2+y**2)),
                (sp.cos(theta), z/sp.sqrt(x**2 + y**2 + z**2))
                ]

        # the cos(theta) dependence encoded by the associated Legendre poly
        expr = sp.assoc_legendre(l, m, sp.cos(theta))

        # the exp(i*m*phi) dependence
        expr *= sp.expand_trig(sp.cos(m*phi)) + sp.I*sp.expand_trig(sp.sin(m*phi))

        # simplifying optimizations
        expr = sp.together(expr.subs(defs)).subs(x**2 + y**2 + z**2, r**2)
        expr = expr.expand().subs([(x/r, xhat), (y/r, yhat), (z/r, zhat)])
        expr = expr.factor().factor(extension=[sp.I]).subs(xhat+sp.I*yhat, xpyhat)
        expr = expr.subs(xhat**2 + yhat**2, 1-zhat**2).factor()

        # and finally add the normalization
        amp = sp.sqrt((2*l+1) / (4*numpy.pi) * sp.factorial(l-m) / sp.factorial(l+m))
        expr *= amp

        return expr



from nbodykit.base.catalog import CatalogSource, column
from nbodykit import CurrentMPIComm
from nbodykit.mpirng import MPIRandomState
import numpy

class RandomCatalog(CatalogSource):
    """
    A CatalogSource that can have columns added via a
    collective random number generator.

    The random number generator stored as :attr:`rng` behaves
    as :class:`numpy.random.RandomState` but generates random
    numbers only on the local rank in a manner independent of
    the number of ranks.

    Parameters
    ----------
    csize : int
        the desired collective size of the Source
    seed : int, optional
        the global seed for the random number generator
    comm : MPI communicator
        the MPI communicator; set automatically if None
    """
    def __repr__(self):
        args = (self.size, self.attrs['seed'])
        return "RandomCatalog(size=%d, seed=%s)" % args

    @CurrentMPIComm.enable
    def __init__(self, csize, seed=None, comm=None):

        self.comm = comm

        # set the seed randomly if it is None
        if seed is None:
            if self.comm.rank == 0:
                seed = numpy.random.randint(0, 4294967295)
            seed = self.comm.bcast(seed)
        self.attrs['seed'] = seed

        # generate the seeds from the global seed
        if csize == 0:
            raise ValueError("no random particles generated!")
        start = comm.rank * csize // comm.size
        end   = (comm.rank + 1) * csize // comm.size
        self._size =  end - start

        self._rng = MPIRandomState(comm, seed=seed, size=self._size)

        # init the base class
        CatalogSource.__init__(self, comm=comm)

    @property
    def rng(self):
        """
        A :class:`MPIRandomState` that behaves as
        :class:`numpy.random.RandomState` but generates random
        numbers in a manner independent of the number of ranks.
        """
        return self._rng

class UniformCatalog(RandomCatalog):
    """
    A CatalogSource that has uniformly-distributed ``Position``
    and ``Velocity`` columns.

    The random numbers generated do not depend on the number of
    available ranks.

    Parameters
    ----------
    nbar : float
        the desired number density of particles in the box
    BoxSize : float, 3-vector
        the size of the box
    seed : int, optional
        the random seed
    comm :
        the MPI communicator
    """
    def __repr__(self):
        args = (self.size, self.attrs['seed'])
        return "UniformCatalog(size=%d, seed=%s)" % args

    @CurrentMPIComm.enable
    def __init__(self, nbar, BoxSize, seed=None, dtype='f8', comm=None):

        self.comm    = comm

        _BoxSize = numpy.empty(3, dtype='f8')
        _BoxSize[:] = BoxSize
        self.attrs['BoxSize'] = _BoxSize

        rng = numpy.random.RandomState(seed)
        N = rng.poisson(nbar * numpy.prod(self.attrs['BoxSize']))
        if N == 0:
            raise ValueError("no uniform particles generated, try increasing `nbar` parameter")
        RandomCatalog.__init__(self, N, seed=seed, comm=comm)

        self._pos = (self.rng.uniform(itemshape=(3,)) * self.attrs['BoxSize']).astype(dtype)
        self._vel = (self.rng.uniform(itemshape=(3,)) * self.attrs['BoxSize'] * 0.01).astype(dtype)

    @column
    def Position(self):
        """
        The position of particles, uniformly distributed in :attr:`BoxSize`
        """
        return self.make_column(self._pos)

    @column
    def Velocity(self):
        """
        The velocity of particles, uniformly distributed in ``0.01 x BoxSize``
        """
        return self.make_column(self._vel)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_allclose, assert_array_equal
import os

setup_logging("debug")

# The test result data (threeptcf_sim_result.dat) is computed with
# Daniel Eisenstein's
# C++ implementation on the same input data set for poles up to l=11;
# We shall agree with it to high precision.
#
# If we need to reproduced these files:
# Nick Hand sent the code and instructions to Yu Feng on Aug-20-2018.

data_dir = os.path.join(os.path.split(os.path.abspath(__file__))[0], 'data')

@MPITest([4])
def test_sim_threeptcf(comm):

    import tempfile

    BoxSize = 400.0

    # load the test data
    filename = os.path.join(data_dir, 'threeptcf_sim_data.dat')
    cat = CSVCatalog(filename, names=['x', 'y', 'z', 'w'], comm=comm)
    cat['Position'] = transform.StackColumns(cat['x'], cat['y'], cat['z'])
    cat['Position'] *= BoxSize

    # r binning
    nbins = 8
    edges = numpy.linspace(0, 200.0, nbins+1)

    # run the algorithm
    ells = list(range(0, 11))
    r = SimulationBox3PCF(cat, ells, edges, BoxSize=BoxSize, weight='w')

    # load the result from file
    truth = numpy.empty((8,8,11))
    with open(os.path.join(data_dir, 'threeptcf_sim_result.dat'), 'r') as ff:
        for line in ff:
            fields = line.split()
            i, j = int(fields[0]), int(fields[1])
            truth[i,j,:] = list(map(float, fields[2:]))
            truth[j,i,:] = truth[i,j,:]

    # test equality
    for i, ell in enumerate(ells):
        x = r.poles['corr_%d' %ell]
        assert_allclose(x * (4*numpy.pi)**2 / (2*ell+1), truth[...,i], rtol=1e-3, err_msg='mismatch for ell=%d' %ell)

    # save to temp file
    filename = 'test-threept-cf.json'
    r.save(filename)
    r2 = SimulationBox3PCF.load(filename, comm=comm)
    assert_array_equal(r.poles.data, r2.poles.data)

    if comm.rank == 0:
        os.remove(filename)


@MPITest([4])
def test_survey_threeptcf(comm):

    import tempfile

    BoxSize = 400.0
    cosmo = cosmology.Planck15

    # load the test data
    filename = os.path.join(data_dir, 'threeptcf_sim_data.dat')
    cat = CSVCatalog(filename, names=['x', 'y', 'z', 'w'], comm=comm)
    cat['Position'] = transform.StackColumns(cat['x'], cat['y'], cat['z'])
    cat['Position'] *= BoxSize

    # place observer at center of box
    cat['Position'] -= 0.5*BoxSize

    # transform to RA/DEC/Z
    # NOTE: we update Position here to get exact same Position as SurveyData3PCF
    cat['RA'], cat['DEC'], cat['Z'] = transform.CartesianToSky(cat['Position'], cosmo)
    cat['Position'] = transform.SkyToCartesian(cat['RA'], cat['DEC'], cat['Z'], cosmo)

    # r binning
    nbins = 8
    edges = numpy.linspace(0, 200.0, nbins+1)

    # run the reference (non-periodic simulation box)
    ells = list(range(0, 11))
    ref = SimulationBox3PCF(cat, ells, edges, BoxSize=BoxSize, weight='w', periodic=False)

    # run the algorithm to test
    r = SurveyData3PCF(cat, ells, edges, cosmo, weight='w', ra='RA', dec='DEC', redshift='Z')

    # test equality between survey result and reference
    for i, ell in enumerate(ells):
        a = r.poles['corr_%d' %ell]
        b = ref.poles['corr_%d' %ell]
        assert_allclose(a, b, rtol=1e-5, err_msg='mismatch for ell=%d' %ell)

    # save to temp file
    filename = 'test-threept-cf.json'
    r.save(filename)
    r2 = SurveyData3PCF.load(filename, comm=comm)
    assert_array_equal(r.poles.data, r2.poles.data)

    if comm.rank == 0:
        os.remove(filename)

@MPITest([1])
def test_sim_threeptcf_pedantic(comm):

    BoxSize = 400.0

    # load the test data
    filename = os.path.join(data_dir, 'threeptcf_sim_data.dat')
    cat = CSVCatalog(filename, names=['x', 'y', 'z', 'w'], comm=comm)
    cat['Position'] = transform.StackColumns(cat['x'], cat['y'], cat['z'])
    cat['Position'] *= BoxSize

    cat = cat[::20]

    # r binning
    nbins = 8
    edges = numpy.linspace(0, 200.0, nbins+1)

    # run the algorithm
    ells = [0, 2, 4, 8]
    r = SimulationBox3PCF(cat, ells, edges, BoxSize=BoxSize, weight='w')
    p_fast = r.run()
    p_pedantic = r.run(pedantic=True)

    # test equality
    for i, ell in enumerate(sorted(ells)):
        x1 = p_fast['corr_%d' %ell]
        x2 = p_pedantic['corr_%d' %ell]
        assert_allclose(x1, x2)

@MPITest([1])
def test_sim_threeptcf_shuffled(comm):

    BoxSize = 400.0

    # load the test data
    filename = os.path.join(data_dir, 'threeptcf_sim_data.dat')
    cat = CSVCatalog(filename, names=['x', 'y', 'z', 'w'], comm=comm)
    cat['Position'] = transform.StackColumns(cat['x'], cat['y'], cat['z'])
    cat['Position'] *= BoxSize

    cat = cat

    # r binning
    nbins = 8
    edges = numpy.linspace(0, 200.0, nbins+1)

    # run the algorithm
    ells = list(range(0, 2))[::-1]
    r = SimulationBox3PCF(cat, ells, edges, BoxSize=BoxSize, weight='w')

    # load the result from file
    truth = numpy.empty((8,8,11))
    with open(os.path.join(data_dir, 'threeptcf_sim_result.dat'), 'r') as ff:
        for line in ff:
            fields = line.split()
            i, j = int(fields[0]), int(fields[1])
            truth[i,j,:] = list(map(float, fields[2:]))
            truth[j,i,:] = truth[i,j,:]

    # test equality
    for i, ell in enumerate(sorted(ells)):
        x = r.poles['corr_%d' %ell]
        assert_allclose(x * (4*numpy.pi)**2 / (2*ell+1), truth[...,i], rtol=1e-3, err_msg='mismatch for ell=%d' %ell)



import numpy
from mpi4py import MPI
import warnings
import functools
import contextlib
import os, sys

def is_structured_array(arr):
    """
    Test if the input array is a structured array
    by testing for `dtype.names`
    """
    if not isinstance(arr, numpy.ndarray) or not hasattr(arr, 'dtype'):
        return False
    return arr.dtype.char ==  'V'

def get_data_bounds(data, comm, selection=None):

    """
    Return the global minimum/maximum of a numpy/dask array along the
    first axis.

    This is computed in chunks to avoid memory errors on large data.

    Parameters
    ----------
    data : numpy.ndarray or dask.array.Array
        the data to find the bounds of
    comm :
        the MPI communicator

    Returns
    -------
    min, max :
        the min/max of ``data``
    """
    import dask.array as da

    # local min/max on this rank
    dmin = numpy.ones(data.shape[1:]) * (numpy.inf)
    dmax = numpy.ones_like(dmin) * (-numpy.inf)

    # max size
    Nlocalmax = max(comm.allgather(len(data)))

    # compute in chunks to avoid memory error
    chunksize = 1024**2 * 8
    for i in range(0, Nlocalmax, chunksize):
        s = slice(i, i + chunksize)

        if len(data) != 0:

            # selection has to be computed many times when data is `large`.
            if selection is not None:
                sel = selection[s]
                if isinstance(selection, da.Array):
                    sel = sel.compute()

            # be sure to use the source to compute
            d = data[s]
            if isinstance(data, da.Array):
                d = d.compute()

            # select
            if selection is not None:
                d = d[sel]

            # update min/max on this rank
            dmin = numpy.min([d.min(axis=0), dmin], axis=0)
            dmax = numpy.max([d.max(axis=0), dmax], axis=0)

    # global min/max across all ranks
    dmin = numpy.asarray(comm.allgather(dmin)).min(axis=0)
    dmax = numpy.asarray(comm.allgather(dmax)).max(axis=0)

    return dmin, dmax

def split_size_3d(s):
    """
    Split `s` into three integers, a, b, c, such
    that a * b * c == s and a <= b <= c

    Parameters
    -----------
    s : int
        integer to split

    Returns
    -------
    a, b, c: int
        integers such that a * b * c == s and a <= b <= c
    """
    a = int(s** 0.3333333) + 1
    d = s
    while a > 1:
        if s % a == 0:
            s = s // a
            break
        a = a - 1
    b = int(s ** 0.5) + 1
    while b > 1:
        if s % b == 0:
            s = s // b
            break
        b = b - 1
    c = s
    return a, b, c

def deprecate(name, alternative, alt_name=None):
    """
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emmitted
    when the function is used.
    """
    alt_name = alt_name or alternative.__name__

    def wrapper(*args, **kwargs):
        warnings.warn("%s is deprecated. Use %s instead" % (name, alt_name),
                      FutureWarning, stacklevel=2)
        return alternative(*args, **kwargs)
    return wrapper

def GatherArray(data, comm, root=0):
    """
    Gather the input data array from all ranks to the specified ``root``.

    This uses `Gatherv`, which avoids mpi4py pickling, and also
    avoids the 2 GB mpi4py limit for bytes using a custom datatype

    Parameters
    ----------
    data : array_like
        the data on each rank to gather
    comm : MPI communicator
        the MPI communicator
    root : int, or Ellipsis
        the rank number to gather the data to. If root is Ellipsis,
        broadcast the result to all ranks.

    Returns
    -------
    recvbuffer : array_like, None
        the gathered data on root, and `None` otherwise
    """
    if not isinstance(data, numpy.ndarray):
        raise ValueError("`data` must by numpy array in GatherArray")

    # need C-contiguous order
    if not data.flags['C_CONTIGUOUS']:
        data = numpy.ascontiguousarray(data)
    local_length = data.shape[0]

    # check dtypes and shapes
    shapes = comm.allgather(data.shape)
    dtypes = comm.allgather(data.dtype)

    # check for structured data
    if dtypes[0].char == 'V':

        # check for structured data mismatch
        names = set(dtypes[0].names)
        if any(set(dt.names) != names for dt in dtypes[1:]):
            raise ValueError("mismatch between data type fields in structured data")

        # check for 'O' data types
        if any(dtypes[0][name] == 'O' for name in dtypes[0].names):
            raise ValueError("object data types ('O') not allowed in structured data in GatherArray")

        # compute the new shape for each rank
        newlength = comm.allreduce(local_length)
        newshape = list(data.shape)
        newshape[0] = newlength

        # the return array
        if root is Ellipsis or comm.rank == root:
            recvbuffer = numpy.empty(newshape, dtype=dtypes[0], order='C')
        else:
            recvbuffer = None

        for name in dtypes[0].names:
            d = GatherArray(data[name], comm, root=root)
            if root is Ellipsis or comm.rank == root:
                recvbuffer[name] = d

        return recvbuffer

    # check for 'O' data types
    if dtypes[0] == 'O':
        raise ValueError("object data types ('O') not allowed in structured data in GatherArray")

    # check for bad dtypes and bad shapes
    if root is Ellipsis or comm.rank == root:
        bad_shape = any(s[1:] != shapes[0][1:] for s in shapes[1:])
        bad_dtype = any(dt != dtypes[0] for dt in dtypes[1:])
    else:
        bad_shape = None; bad_dtype = None

    bad_shape, bad_dtype = comm.bcast((bad_shape, bad_dtype))

    if bad_shape:
        raise ValueError("mismatch between shape[1:] across ranks in GatherArray")
    if bad_dtype:
        raise ValueError("mismatch between dtypes across ranks in GatherArray")

    shape = data.shape
    dtype = data.dtype

    # setup the custom dtype
    duplicity = numpy.product(numpy.array(shape[1:], 'intp'))
    itemsize = duplicity * dtype.itemsize
    dt = MPI.BYTE.Create_contiguous(itemsize)
    dt.Commit()

    # compute the new shape for each rank
    newlength = comm.allreduce(local_length)
    newshape = list(shape)
    newshape[0] = newlength

    # the return array
    if root is Ellipsis or comm.rank == root:
        recvbuffer = numpy.empty(newshape, dtype=dtype, order='C')
    else:
        recvbuffer = None

    # the recv counts
    counts = comm.allgather(local_length)
    counts = numpy.array(counts, order='C')

    # the recv offsets
    offsets = numpy.zeros_like(counts, order='C')
    offsets[1:] = counts.cumsum()[:-1]

    # gather to root
    if root is Ellipsis:
        comm.Allgatherv([data, dt], [recvbuffer, (counts, offsets), dt])
    else:
        comm.Gatherv([data, dt], [recvbuffer, (counts, offsets), dt], root=root)

    dt.Free()

    return recvbuffer

def ScatterArray(data, comm, root=0, counts=None):
    """
    Scatter the input data array across all ranks, assuming `data` is
    initially only on `root` (and `None` on other ranks).

    This uses ``Scatterv``, which avoids mpi4py pickling, and also
    avoids the 2 GB mpi4py limit for bytes using a custom datatype

    Parameters
    ----------
    data : array_like or None
        on `root`, this gives the data to split and scatter
    comm : MPI communicator
        the MPI communicator
    root : int
        the rank number that initially has the data
    counts : list of int
        list of the lengths of data to send to each rank

    Returns
    -------
    recvbuffer : array_like
        the chunk of `data` that each rank gets
    """
    import logging

    if counts is not None:
        counts = numpy.asarray(counts, order='C')
        if len(counts) != comm.size:
            raise ValueError("counts array has wrong length!")

    # check for bad input
    if comm.rank == root:
        bad_input = not isinstance(data, numpy.ndarray)
    else:
        bad_input = None
    bad_input = comm.bcast(bad_input)
    if bad_input:
        raise ValueError("`data` must by numpy array on root in ScatterArray")

    if comm.rank == 0:
        # need C-contiguous order
        if not data.flags['C_CONTIGUOUS']:
            data = numpy.ascontiguousarray(data)
        shape_and_dtype = (data.shape, data.dtype)
    else:
        shape_and_dtype = None

    # each rank needs shape/dtype of input data
    shape, dtype = comm.bcast(shape_and_dtype)

    # object dtype is not supported
    fail = False
    if dtype.char == 'V':
         fail = any(dtype[name] == 'O' for name in dtype.names)
    else:
        fail = dtype == 'O'
    if fail:
        raise ValueError("'object' data type not supported in ScatterArray; please specify specific data type")

    # initialize empty data on non-root ranks
    if comm.rank != root:
        np_dtype = numpy.dtype((dtype, shape[1:]))
        data = numpy.empty(0, dtype=np_dtype)

    # setup the custom dtype
    duplicity = numpy.product(numpy.array(shape[1:], 'intp'))
    itemsize = duplicity * dtype.itemsize
    dt = MPI.BYTE.Create_contiguous(itemsize)
    dt.Commit()

    # compute the new shape for each rank
    newshape = list(shape)

    if counts is None:
        newlength = shape[0] // comm.size
        if comm.rank < shape[0] % comm.size:
            newlength += 1
        newshape[0] = newlength
    else:
        if counts.sum() != shape[0]:
            raise ValueError("the sum of the `counts` array needs to be equal to data length")
        newshape[0] = counts[comm.rank]

    # the return array
    recvbuffer = numpy.empty(newshape, dtype=dtype, order='C')

    # the send counts, if not provided
    if counts is None:
        counts = comm.allgather(newlength)
        counts = numpy.array(counts, order='C')

    # the send offsets
    offsets = numpy.zeros_like(counts, order='C')
    offsets[1:] = counts.cumsum()[:-1]

    # do the scatter
    comm.Barrier()
    comm.Scatterv([data, (counts, offsets), dt], [recvbuffer, dt])
    dt.Free()
    return recvbuffer

def FrontPadArray(array, front, comm):
    """ Padding an array in the front with items before this rank.

    """
    N = numpy.array(comm.allgather(len(array)), dtype='intp')
    offsets = numpy.cumsum(numpy.concatenate([[0], N], axis=0))
    mystart = offsets[comm.rank] - front
    torecv = (offsets[:-1] + N) - mystart

    torecv[torecv < 0] = 0 # before mystart
    torecv[torecv > front] = 0 # no more than needed
    torecv[torecv > N] = N[torecv > N] # fully enclosed

    if comm.allreduce(torecv.sum() != front, MPI.LOR):
        raise ValueError("cannot work out a plan to padd items. Some front values are too large. %d %d"
            % (torecv.sum(), front))

    tosend = comm.alltoall(torecv)
    sendbuf = [ array[-items:] if items > 0 else array[0:0] for i, items in enumerate(tosend)]
    recvbuf = comm.alltoall(sendbuf)
    return numpy.concatenate(list(recvbuf) + [array], axis=0)

def attrs_to_dict(obj, prefix):
    if not hasattr(obj, 'attrs'):
        return {}

    d = {}
    for key, value in obj.attrs.items():
        d[prefix + key] = value
    return d

import json
from astropy.units import Quantity, Unit
from nbodykit.cosmology import Cosmology

class JSONEncoder(json.JSONEncoder):
    """
    A subclass of :class:`json.JSONEncoder` that can also handle numpy arrays,
    complex values, and :class:`astropy.units.Quantity` objects.
    """
    def default(self, obj):

        # Cosmology object
        if isinstance(obj, Cosmology):
            d = {}
            d['__cosmo__'] = obj.pars.copy()
            return d

        # astropy quantity
        if isinstance(obj, Quantity):

            d = {}
            d['__unit__'] = str(obj.unit)

            value = obj.value
            if obj.size > 1:
                d['__dtype__'] = value.dtype.str if value.dtype.names is None else value.dtype.descr
                d['__shape__'] = value.shape
                value = value.tolist()

            d['__data__'] = value
            return d

        # complex values
        elif isinstance(obj, complex):
            return {'__complex__': [obj.real, obj.imag ]}

        # numpy arrays
        elif isinstance(obj, numpy.ndarray):
            value = obj
            dtype = obj.dtype
            d = {
                '__dtype__' :
                    dtype.str if dtype.names is None else dtype.descr,
                '__shape__' : value.shape,
                '__data__': value.tolist(),
            }
            return d
        # explicity convert numpy data types to python types
        # see: https://bugs.python.org/issue24313
        elif isinstance(obj, numpy.floating):
            return float(obj)
        elif isinstance(obj, numpy.integer):
            return int(obj)

        return json.JSONEncoder.default(self, obj)

class JSONDecoder(json.JSONDecoder):
    """
    A subclass of :class:`json.JSONDecoder` that can also handle numpy arrays,
    complex values, and :class:`astropy.units.Quantity` objects.
    """
    @staticmethod
    def hook(value):
        def fixdtype(dtype):
            if isinstance(dtype, list):
                true_dtype = []
                for field in dtype:
                    if len(field) == 3:
                        true_dtype.append((str(field[0]), str(field[1]), field[2]))
                    if len(field) == 2:
                        true_dtype.append((str(field[0]), str(field[1])))
                return true_dtype
            return dtype

        def fixdata(data, N, dtype):
            if not isinstance(dtype, list):
                return data

            # for structured array,
            # the last dimension shall be a tuple
            if N > 0:
                return [fixdata(i, N - 1, dtype) for i in data]
            else:
                assert len(data) == len(dtype)
                return tuple(data)

        d = None
        if '__dtype__' in value:
            dtype = fixdtype(value['__dtype__'])
            shape = value['__shape__']
            a = fixdata(value['__data__'], len(shape), dtype)
            d = numpy.array(a, dtype=dtype)

        if '__unit__' in value:
            if d is None:
                d = value['__data__']
            d = Quantity(d, Unit(value['__unit__']))

        if '__cosmo__' in value:
            d = Cosmology.from_dict(value['__cosmo__'])

        if d is not None:
            return d

        if '__complex__' in value:
            real, imag = value['__complex__']
            return real + 1j * imag

        return value

    def __init__(self, *args, **kwargs):
        kwargs['object_hook'] = JSONDecoder.hook
        json.JSONDecoder.__init__(self, *args, **kwargs)

def timer(start, end):
    """
    Utility function to return a string representing the elapsed time,
    as computed from the input start and end times

    Parameters
    ----------
    start : int
        the start time in seconds
    end : int
        the end time in seconds

    Returns
    -------
    str :
        the elapsed time as a string, using the format `hours:minutes:seconds`
    """
    hours, rem = divmod(end-start, 3600)
    minutes, seconds = divmod(rem, 60)
    return "{:0>2}:{:0>2}:{:05.2f}".format(int(hours),int(minutes),seconds)

@contextlib.contextmanager
def captured_output(comm, root=0):
    """
    Re-direct stdout and stderr to null for every rank but ``root``
    """
    # keep output on root
    if root is not None and comm.rank == root:
        yield sys.stdout, sys.stderr
    else:
        from six.moves import StringIO
        from nbodykit.extern.wurlitzer import sys_pipes

        # redirect stdout and stderr
        old_stdout, sys.stdout = sys.stdout, StringIO()
        old_stderr, sys.stderr = sys.stderr, StringIO()
        try:
            with sys_pipes() as (out, err):
                yield out, err
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from nbodykit.utils import ScatterArray, GatherArray

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_fibercolls(comm):

    from scipy.spatial.distance import pdist, squareform 


    N = 10000

    # generate the initial data
    numpy.random.seed(42)
    if comm.rank == 0:
        ra = 10.*numpy.random.random(size=N)
        dec = 5.*numpy.random.random(size=N) - 5.0
    else:
        ra = None
        dec = None

    ra = ScatterArray(ra, comm)
    dec = ScatterArray(dec, comm)

    # compute the fiber collisions
    r = FiberCollisions(ra, dec, degrees=True, seed=42, comm=comm)
    rad = r._collision_radius_rad

    #  gather collided and position to root
    idx = GatherArray(r.labels['Collided'].compute().astype(bool), comm, root=0)
    pos = GatherArray(r.source['Position'].compute(), comm, root=0)

    # manually compute distances and check on root
    if comm.rank == 0:
        dists = squareform(pdist(pos, metric='euclidean'))
        numpy.fill_diagonal(dists, numpy.inf) # ignore self pairs

        # no objects in clean sample (Collided==0) should be within
        # the collision radius of any other objects in the sample
        clean_dists = dists[~idx, ~idx]
        assert (clean_dists <= rad).sum() == 0, "objects in 'clean' sample within collision radius!"

        # the collided objects must collided with at least 
        # one object in the clean sample
        ncolls_per = (dists[idx] <= rad).sum(axis=-1)
        assert (ncolls_per >= 1).all(), "objects in 'collided' sample that do not collide with any objects!"




from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import os
import pytest

setup_logging()

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_sim_data(seed, comm):
    return UniformCatalog(nbar=3e-6, BoxSize=512., seed=seed, comm=comm)

def generate_survey_data(seed, comm):
    cosmo = cosmology.Planck15
    s = RandomCatalog(1000, seed=seed, comm=comm)

    # ra, dec, z
    s['Redshift'] = s.rng.normal(loc=0.5, scale=0.1)
    s['RA'] = s.rng.uniform(low=110, high=260)
    s['DEC'] = s.rng.uniform(low=-3.6, high=60.)

    # position
    s['Position'] = transform.SkyToCartesian(s['RA'], s['DEC'], s['Redshift'], cosmo=cosmo)

    return s

def reference_sim_paircount(pos1, redges, pimax, boxsize, pos2=None, los=2):
    """Reference pair counting via halotools"""
    from halotools.mock_observables.pair_counters import npairs_xy_z

    # reorder the axes
    axes_order = [i for i in [0,1,2] if i != los] + [los]
    pos1 = pos1[:,axes_order]

    # the pi bins
    pibins = numpy.linspace(0, pimax, pimax+1)

    if pos2 is None:
        pos2 = pos1
    else:
        pos2 = pos2[:,axes_order]
    r = npairs_xy_z(pos1, pos2, redges, pibins, period = boxsize)
    return numpy.diff(numpy.diff(r, axis=0), axis=1)

def reference_survey_paircount(pos1, w1, rp_bins, pimax, pos2=None, w2=None, los=2):
    """Reference pair counting via kdcount"""
    from kdcount import KDTree, correlate

    pi_bins = numpy.linspace(0, pimax, int(pimax)+1)
    tree1 = correlate.points(pos1, boxsize=None, weights=w1).tree.root
    if pos2 is None:
        pos2 = pos1; w2 = w1
    tree2 = correlate.points(pos2, boxsize=None, weights=w2).tree.root

    if w1 is None:
        w1 = numpy.ones_like(pos1)
    if w2 is None:
        w2 = numpy.ones_like(pos2)

    # find all pairs
    r, i, j = tree1.enum(tree2, (pimax**2 + rp_bins[-1]**2)**0.5)

    def compute_rp_pi(r, i, j):
        r1 = pos1[i]
        r2 = pos2[j]
        center = 0.5 * (r1 + r2)
        dr = r1 - r2
        dot = numpy.einsum('ij, ij->i', dr, center)
        dot2 = dot ** 2
        center2 = numpy.einsum('ij, ij->i', center, center)
        los2 = dot2 / center2
        dr2 = numpy.einsum('ij, ij->i', dr, dr)
        x2 = numpy.abs(dr2 - los2)
        return x2 ** 0.5, los2 ** 0.5

    # compute rp, pi distances
    rp, pi = compute_rp_pi(r, i, j)

    # digitize
    rp_dig = numpy.digitize(rp, rp_bins)
    pi_dig = numpy.digitize(pi, pi_bins)
    shape = (len(rp_bins)+1,len(pi_bins)+1)
    multi_index = numpy.ravel_multi_index([rp_dig, pi_dig], shape)

    # initialize the return arrays
    npairs = numpy.zeros(shape, dtype='i8')
    rpavg = numpy.zeros(shape, dtype='f8')
    wnpairs = numpy.zeros(shape, dtype='f8')

    # mean rp values
    rpavg.flat += numpy.bincount(multi_index, weights=rp, minlength=rpavg.size)
    rpavg = rpavg[1:-1,1:-1]

    # count the pairs
    npairs.flat += numpy.bincount(multi_index, minlength=npairs.size)
    npairs = npairs[1:-1,1:-1]

    wnpairs.flat += numpy.bincount(multi_index, weights=w1[i]*w2[j], minlength=wnpairs.size)
    wnpairs = wnpairs[1:-1,1:-1]

    return npairs, numpy.nan_to_num(rpavg/npairs), numpy.nan_to_num(wnpairs)


@MPITest([1, 3])
def test_sim_periodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    pimax = 40.
    r = SimulationBoxPairCount('projected', source, redges, periodic=True, weight='Weight', pimax=pimax)

    pos = gather_data(source, "Position")

    # verify with halotools
    npairs = reference_sim_paircount(pos, redges, pimax, source.attrs['BoxSize'])
    assert_allclose(npairs, r.pairs['npairs'])

@MPITest([4])
def test_sim_diff_los(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)

    # add some weights b/w 0 and 1
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the weighted paircount
    pimax = 100.
    r = SimulationBoxPairCount('projected', source, redges, periodic=True, weight='Weight', pimax=pimax, los='x')

    # verify with halotools
    pos = gather_data(source, "Position")
    npairs = reference_sim_paircount(pos, redges, pimax, source.attrs['BoxSize'], los=0)
    assert_allclose(npairs, r.pairs['npairs'])

@MPITest([1, 3])
def test_sim_nonperiodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)
    source['Weight'] = numpy.random.random(size=len(source))

    # make the bin edges
    redges = numpy.linspace(10, 250., 10)

    # do the weighted paircount
    pimax = 50.
    r = SimulationBoxPairCount('projected', source, redges, periodic=False, weight='Weight', pimax=pimax)

    # verify with halotools
    pos = gather_data(source, "Position")
    npairs = reference_sim_paircount(pos, redges, pimax, None)
    assert_allclose(npairs, r.pairs['npairs'])


@MPITest([1, 3])
def test_sim_periodic_cross(comm):

    # generate data
    first = generate_sim_data(seed=42, comm=comm)
    second = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(10, 40., 10)

    # do the paircount
    pimax = 10.
    r = SimulationBoxPairCount('projected', first, redges, second=second, periodic=True, pimax=pimax)

    pos1 = gather_data(first, "Position")
    pos2 = gather_data(second, "Position")

    # verify with halotools
    npairs = reference_sim_paircount(pos1, redges, pimax, first.attrs['BoxSize'], pos2=pos2)
    assert_allclose(npairs, r.pairs['npairs'])

@MPITest([1, 4])
def test_survey_auto(comm):
    cosmo = cosmology.Planck15

    # random particles
    source = generate_survey_data(seed=42, comm=comm)
    source['Weight'] = source.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 1000., 10)

    # do the weighted paircount
    pimax = 500.
    r = SurveyDataPairCount('projected', source, redges, cosmo, weight='Weight', pimax=pimax)

    pos = gather_data(source, 'Position')
    w = gather_data(source, 'Weight')

    # verify with kdcount
    npairs, ravg, wnpairs = reference_survey_paircount(pos, w, redges, pimax)
    assert_allclose(ravg, r.pairs['rp'], rtol=1e-5)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wnpairs, r.pairs['wnpairs'])

@MPITest([1, 4])
def test_survey_cross(comm):
    cosmo = cosmology.Planck15

    # random particles
    first = generate_survey_data(seed=42, comm=comm)
    first['Weight'] = first.rng.uniform()
    second = generate_survey_data(seed=84, comm=comm)
    second['Weight'] = second.rng.uniform()

    # make the bin edges
    redges = numpy.linspace(10, 150, 10)

    # do the paircount
    pimax = 50.
    r = SurveyDataPairCount('projected', first, redges, cosmo, second=second, pimax=pimax)

    pos1 = gather_data(first, 'Position')
    pos2 = gather_data(second, 'Position')
    w1 = gather_data(first, 'Weight')
    w2 = gather_data(second, 'Weight')

    # verify with kdcount
    npairs, ravg, wnpairs = reference_survey_paircount(pos1, w1, redges, pimax, pos2=pos2, w2=w2)
    assert_allclose(ravg, r.pairs['rp'], rtol=1e-3)
    assert_allclose(npairs, r.pairs['npairs'])
    assert_allclose(wnpairs, r.pairs['wnpairs'])

    # test save
    r.save('paircount-test.json')
    r2 = SurveyDataPairCount.load('paircount-test.json', comm=comm)
    assert_array_equal(r.pairs.data, r2.pairs.data)

    if comm.rank == 0: os.remove('paircount-test.json')

@MPITest([1])
def test_missing_pimax(comm):

    # generate data
    source = generate_sim_data(seed=42, comm=comm)
    redges = numpy.linspace(10, 150, 10)

    # missing pimax
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('projected', source, redges)

    # wrong mode
    with pytest.raises(ValueError):
        r = SimulationBoxPairCount('1d', source, redges, pimax=10.)

@MPITest([1])
def test_bad_pimax(comm):

    # generate data
    source = generate_sim_data(seed=42, comm=comm)

    # pimax must be at least 1
    with pytest.raises(ValueError):
        redges = numpy.linspace(10, 150, 10)
        r = SimulationBoxPairCount('projected', source, redges, pimax=0.5)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

import shutil
from numpy.testing import assert_array_equal, assert_allclose

setup_logging()

@MPITest([1,4])
def test_bigfile_grid(comm):

    import tempfile

    cosmo = cosmology.Planck15

    # input linear mesh
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, BoxSize=512, Nmesh=32, seed=42, comm=comm)

    real = source.compute(mode='real')
    complex = source.compute(mode="complex")

    # and save to tmp directory
    if comm.rank == 0:
        output = tempfile.mkdtemp()
    else:
        output = None
    output = comm.bcast(output)

    source.save(output, dataset='Field')

    # now load it and paint to the algorithm's ParticleMesh
    source = BigFileMesh(path=output, dataset='Field', comm=comm)
    loaded_real = source.compute()

    # compare to direct algorithm result
    assert_array_equal(real, loaded_real)

    source.save(output, dataset='FieldC', mode='complex')

    # now load it and paint to the algorithm's ParticleMesh
    source = BigFileMesh(path=output, dataset='FieldC', comm=comm)
    loaded_real = source.compute(mode="complex")

    # compare to direct algorithm result
    assert_allclose(complex, loaded_real, atol=1e-7)
    if comm.rank == 0:
        shutil.rmtree(output)



from .binary import BinaryFile
import numpy
from six import string_types
from . import tools
import warnings

DefaultColumnDefs = [
    ('Position', ('auto', 3), 'all', ),
    ('GadgetVelocity',  ('auto', 3), 'all', ),
    ('ID', 'auto', 'all', ),
    ('Mass', 'auto', None, ),
    ('InternalEnergy', 'auto', (0, ), ),
    ('Density', 'auto', (0, ), ),
    ('SmoothingLength', 'auto', (0,) ),
    ]

DefaultHeaderDtype = [
        ('Npart', ('u4', 6)),
        ('Massarr', ('f8', 6)),
        ('Time', ('f8')),
        ('Redshift', ('f8')),
        ('FlagSfr', ('i4')),
        ('FlagFeedback', ('i4')),
        ('Nall', ('u4', 6)),
        ('FlagCooling', ('i4')),
        ('NumFiles', ('i4')),
        ('BoxSize', ('f8')),
        ('Omega0', ('f8')),
        ('OmegaLambda', ('f8')),
        ('HubbleParam', ('f8')),
        ('FlagAge', ('i4')),
        ('FlagMetals', ('i4')),
        ('NallHW', ('u4', 6)),
        ('flag_entr_ics', ('i4')),
    ]

class Gadget1File(BinaryFile):
    """
    Read snapshot binary files from Volkers Gadget 1/2/3 simulations.

    These files are stored column-wise with a format, with a
    header of size 28 bytes to begin the file.

    The columns are:

    * Position : 'f4', 'f8' precision
        the position data, usually in Kpc/h units.
    * GadgetVelocity : 'f4', 'f8' precision
        the Gadget 1 velocity, sqrt(a)**-1 v_p. in km/s
    * ID : 'i8'/'i4' precision
        integers specfiying the particle ID

    Parameters
    ----------
    path : str
        the path to the binary file to load
    columndefs : list
        a list of triplets (columnname, element_dtype, particle_types)
    ptype : int
        type of particle of interest.
    hdtype : list, dtype
        dtype of the header; must define Massarr and Npart

    References
    ----------
    https://wwwmpa.mpa-garching.mpg.de/gadget/users-guide.pdf
    """
    def __init__(self, path, columndefs=DefaultColumnDefs,
                hdtype=DefaultHeaderDtype, ptype=1):

        if ptype not in [0, 1, 2, 3, 4, 5]:
            raise ValueError("ptype shall be 0 ~ 5.")

        hdtype = numpy.dtype(hdtype)
        hdtype_padded = numpy.dtype([
                                     ('f77', 'i4'),
                                     ('header', hdtype),
                                     ('padding', ('u1', 256 - hdtype.itemsize))])
        header = numpy.fromfile(path, dtype=hdtype_padded, count=1)[0]['header']

        attrs = {}

        for key in header.dtype.names:
            attrs[key] = header[key].copy()

        self.attrs = attrs

        self.file_header = header
        self.header_mass = header['Massarr'][ptype]
        self.ptype = ptype

        dtype = []
        defs = []

        with open(path, 'r') as ff:
            offsets = {}
            ptr = 256 + 4 + 4
            for column, spec, ptypes in columndefs:
                if not isinstance(spec, tuple):
                    spec = spec, ()
                if len(spec) == 1:
                    spec = spec[0], ()

                if ptypes == 'all':
                    ptypes = [0, 1, 2, 3, 4, 5]
                elif column == 'Mass':
                    ptypes = (header['Massarr'] == 0).nonzero()[0]

                blocksize = 0
                reloffset = 0
                N = 0
                for i in ptypes:
                    if i == ptype:
                        reloffset = N
                    N += int(header['Npart'][i])

                if N != 0: # block exists
                    ff.seek(ptr, 0)
                    a = numpy.fromfile(ff, dtype='i4', count=1)[0]
                    ptr += 4

                    itemsize = a // N # compute precision from blocksize

                    blocksize = N * itemsize

                    offsets[column] = ptr + reloffset * itemsize

                    ptr += a

                    ff.seek(ptr, 0)
                    b = numpy.fromfile(ff, dtype='i4', count=1)[0]
                    ptr += 4

                    if a != b or b != blocksize:
                        raise IOError("F77 unformatted meta data for `%s` disagrees with true size: starting = %d truth = %d ending = %d" % (column, a, blocksize, b))

                    itemshape = numpy.prod(spec[1])
                    prec = itemsize // itemshape 
                else:
                    offsets[column] = ptr
                    warnings.warn("Cannot decide the item size of `%s`, assuming 4 bytes." % (column))
                    prec = None

                if spec[0] == 'auto':
                    if column == 'ID':
                        mapping = {8:'i8', 4:'i4', None:'i4'}
                    else:
                        mapping = {8:'f8', 4:'f4', None:'f8'}

                    spec = mapping[prec], spec[1]

                if column == "Mass" or ptype in ptypes:
                    dtype.append((column, spec))

                defs.append((column, spec, ptypes))

        dtype = numpy.dtype(dtype)

        self.defs = defs

        BinaryFile.__init__(self, path, dtype=dtype, header_size=256+4+4, offsets=offsets, size=int(header['Npart'][ptype]))


    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range

        'start' and 'stop' should be between 0 and :attr:`size`,
        which is the total size of the binary file (in particles)

        Parameters
        ----------
        columns : str, list of str
            the name of the column(s) to return
        start : int
            the row integer to start reading at
        stop : int
            the row integer to stop reading at
        step : int, optional
            the step size to use when reading; default is 1

        Returns
        -------
        numpy.array
            structured array holding the requested columns over
            the specified range of rows
        """
        if isinstance(columns, string_types): columns = [columns]

        if stop > self.size or start > self.size or start < 0 or stop < 0:
            raise IndexError("start : %d stop %d beyond size of data set %d"
                % (start, stop, self.size))

        dt = [(col, self.dtype[col]) for col in columns]
        toret = numpy.empty(tools.get_slice_size(start, stop, step), dtype=dt)

        with open(self.path, 'rb') as ff:

            for col in columns:
                offset = self.offsets[col]
                dtype = self.dtype[col]
                if col == 'Mass' and self.header_mass != 0:
                    toret[col][:] = self.header_mass
                else:
                    ff.seek(offset, 0)
                    ff.seek(start * dtype.itemsize, 1)
                    toret[col][:] = numpy.fromfile(ff, count=stop-start, dtype=dtype)[::step]

        return toret



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_array_equal, assert_allclose
import pytest

setup_logging()

@MPITest([1])
def test_boxsize_nmesh(comm):

    # the catalog
    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)

    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    # this should work (infer BoxSize)
    mesh = cat.to_mesh(Nmesh=32)

    # this should not work (no Nmesh to infer)
    with pytest.raises(ValueError):
        mesh = cat.to_mesh()

    # this not should work
    cat.attrs['data.BoxSize'] *= 10.
    with pytest.raises(ValueError):
        mesh = cat.to_mesh(Nmesh=32)

@MPITest([1, 4])
def test_getitem(comm):

    # the catalog
    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)
    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    # the mesh
    mesh = cat.to_mesh(Nmesh=32, BoxSize=512)

    for source, name in zip([source1, source2], ['data', 'randoms']):
        submesh = mesh[name] # should be equal to source
        assert submesh.source is cat[name]

@MPITest([1, 4])
def test_compute(comm):

    # the catalog
    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)
    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    # the meshes
    mesh = cat.to_mesh(Nmesh=32, BoxSize=512)
    mesh1 = source1.to_mesh(Nmesh=32, BoxSize=512)
    mesh2 = source2.to_mesh(Nmesh=32, BoxSize=512)

    # paint
    real1 = mesh1.to_real_field()
    real2 = mesh2.to_real_field()

    # un-normalize real1 and real2
    real1[:] *= real1.attrs['num_per_cell']
    real2[:] *= real2.attrs['num_per_cell']
    norm = real1.attrs['num_per_cell'] + real2.attrs['num_per_cell']

    # the combined density field
    combined = mesh.to_real_field()

    # must be the same
    assert_allclose(combined.value, (real1.value + real2.value)/norm, atol=1e-5)

@MPITest([1, 4])
def test_paint_interlaced(comm):

    # the test case fails only if there is enough particles to trigger
    # the second loop of the interlaced painter; these parameters will do it.

    # the catalog
    source1 = UniformCatalog(nbar=1e-0, BoxSize=111, seed=111, comm=comm)
    source2 = UniformCatalog(nbar=1e-0, BoxSize=111, seed=111, comm=comm)
    source1['Weight'] = 1.0
    source2['Weight'] = 0.1
    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    # the meshes
    mesh = cat.to_mesh(Nmesh=32, interlaced=True)
    mesh1 = source1.to_mesh(Nmesh=32, interlaced=True)
    mesh2 = source2.to_mesh(Nmesh=32, interlaced=True)

    # paint
    real1 = mesh1.to_real_field()
    real2 = mesh2.to_real_field()
    assert_allclose(real1.cmean(), 1.0)
    assert_allclose(real2.cmean(), 1.0)

    # un-normalize real1 and real2
    real1[:] *= real1.attrs['num_per_cell']
    real2[:] *= real2.attrs['num_per_cell']
    norm = real1.attrs['num_per_cell'] + real2.attrs['num_per_cell']

    # the combined density field
    #combined = mesh.to_real_field()
    combined = mesh.compute()

    assert_allclose(combined.cmean(), 1.0)
    # must be the same
    assert_allclose(combined.value, (real1.value + real2.value)/norm, atol=1e-5)



import numpy
import numbers
from mpi4py import MPI

from pmesh.pm import RealField, ComplexField
from nbodykit.meshtools import SlabIterator
from nbodykit.utils import GatherArray, ScatterArray
from nbodykit.mpirng import MPIRandomState
import mpsort

def gaussian_complex_fields(pm, linear_power, seed,
            unitary_amplitude=False, inverted_phase=False,
            compute_displacement=False, logger=None):
    r"""
    Make a Gaussian realization of a overdensity field, :math:`\delta(x)`.

    If specified, also compute the corresponding 1st order Lagrangian
    displacement field (Zel'dovich approximation) :math:`\psi(x)`,
    which is related to the linear velocity field via:

    .. math::

        v(x) = f a H \psi(x)

    Notes
    -----
    This computes the overdensity field using the following steps:

    #. Generate complex variates with unity variance
    #. Scale the Fourier field by :math:`(P(k) / V)^{1/2}`

    After step 2, the complex field has unity variance. This
    is equivalent to generating real-space normal variates
    with mean and unity variance, calling r2c() and dividing by :math:`N^3`
    since the variance of the complex FFT (with no additional normalization)
    is :math:`N^3 \times \sigma^2_\mathrm{real}`.

    Furthermore, the power spectrum is defined as V * variance.
    So a normalization factor of 1 / V shows up in step 2,
    cancels this factor such that the power spectrum is P(k).

    The linear displacement field is computed as:

    .. math::

        \psi_i(k) = i \frac{k_i}{k^2} \delta(k)

    .. note::

        To recover the linear velocity in proper units, i.e., km/s,
        from the linear displacement, an additional factor of
        :math:`f \times a \times H(a)` is required

    Parameters
    ----------
    pm : pmesh.pm.ParticleMesh
        the mesh object
    linear_power : callable
        a function taking wavenumber as its only argument, which returns
        the linear power spectrum
    seed : int
        the random seed used to generate the random field
    compute_displacement : bool, optional
        if ``True``, also return the linear Zel'dovich displacement field;
        default is ``False``
    unitary_amplitude : bool, optional
        if ``True``, the seed gaussian has unitary_amplitude.
    inverted_phase: bool, optional
        if ``True``, the phase of the seed gaussian is inverted

    Returns
    -------
    delta_k : ComplexField
        the real-space Gaussian overdensity field
    disp_k : ComplexField or ``None``
        if requested, the Gaussian displacement field
    """
    if not isinstance(seed, numbers.Integral):
        raise ValueError("the seed used to generate the linear field must be an integer")

    if logger and pm.comm.rank == 0:
        logger.info("Generating whitenoise")

    # use pmesh to generate random complex white noise field (done in parallel)
    # variance of complex field is unity
    # multiply by P(k)**0.5 to get desired variance
    delta_k = pm.generate_whitenoise(seed, type='untransposedcomplex', unitary=unitary_amplitude)

    if logger and pm.comm.rank == 0:
        logger.info("Write noise generated")

    if inverted_phase: delta_k[...] *= -1

    # initialize the displacement fields for (x,y,z)
    if compute_displacement:
        disp_k = [pm.create(type='untransposedcomplex') for i in range(delta_k.ndim)]
        for i in range(delta_k.ndim): disp_k[i][:] = 1j
    else:
        disp_k = None

    # volume factor needed for normalization
    norm = 1.0 / pm.BoxSize.prod()

    # iterate in slabs over fields
    slabs = [delta_k.slabs.x, delta_k.slabs]
    if compute_displacement:
        slabs += [d.slabs for d in disp_k]

    # loop over the mesh, slab by slab
    for islabs in zip(*slabs):
        kslab, delta_slab = islabs[:2] # the k arrays and delta slab

        # the square of the norm of k on the mesh
        k2 = sum(kk**2 for kk in kslab)
        zero_idx = k2 == 0.

        k2[zero_idx] = 1. # avoid dividing by zero

        # the linear power (function of k)
        power = linear_power((k2**0.5).flatten())

        # multiply complex field by sqrt of power
        delta_slab[...].flat *= (power*norm)**0.5

        # set k == 0 to zero (zero config-space mean)
        delta_slab[zero_idx] = 0.

        # compute the displacement
        if compute_displacement:

            # ignore division where k==0 and set to 0
            with numpy.errstate(invalid='ignore', divide='ignore'):
                for i in range(delta_k.ndim):
                    disp_slab = islabs[2+i]
                    disp_slab[...] *= kslab[i] / k2 * delta_slab[...]
                    disp_slab[zero_idx] = 0. # no bulk displacement

    if logger and pm.comm.rank == 0:
        logger.info("Displacement computed in fourier space")

    # return Fourier-space density and displacement (which could be None)
    return delta_k, disp_k


def gaussian_real_fields(pm, linear_power, seed,
                unitary_amplitude=False,
                inverted_phase=False, compute_displacement=False, logger=None):
    r"""
    Make a Gaussian realization of a overdensity field in
    real-space :math:`\delta(x)`.

    If specified, also compute the corresponding linear Zel'dovich
    displacement field :math:`\psi(x)`, which is related to the
    linear velocity field via:

    Notes
    -----
    See the docstring for :func:`gaussian_complex_fields` for the
    steps involved in generating the fields.

    Parameters
    ----------
    pm : pmesh.pm.ParticleMesh
        the mesh object
    linear_power : callable
        a function taking wavenumber as its only argument, which returns
        the linear power spectrum
    seed : int
        the random seed used to generate the random field
    compute_displacement : bool, optional
        if ``True``, also return the linear Zel'dovich displacement field;
        default is ``False``
    unitary_amplitude : bool, optional
        if ``True``, the seed gaussian has unitary_amplitude.
    inverted_phase: bool, optional
        if ``True``, the phase of the seed gaussian is inverted

    Returns
    -------
    delta : RealField
        the real-space Gaussian overdensity field
    disp : RealField or ``None``
        if requested, the Gaussian displacement field
    """
    # make fourier fields
    delta_k, disp_k = gaussian_complex_fields(pm, linear_power, seed,
                            inverted_phase=inverted_phase,
                            unitary_amplitude=unitary_amplitude,
                            compute_displacement=compute_displacement,
                            logger=logger)

    # FFT the density to real-space
    delta = delta_k.c2r()

    std = (delta ** 2).cmean() ** 0.5

    if logger and pm.comm.rank == 0:
        logger.info("Overdensity computed in configuration space: std = %s" % str(std))

    # FFT the velocity back to real space
    if compute_displacement:
        disp = [disp_k[i].c2r() for i in range(delta.ndim)]

        std = [(disp[i] ** 2).cmean() ** 0.5 for i in range(delta.ndim)]

        if logger and pm.comm.rank == 0:
            logger.info("Displacement computed in configuration space: std = %s" % str(std))
    else:
        disp = None

    # return density and displacement (which could be None)
    return delta, disp


def lognormal_transform(density, bias=1.):
    r"""
    Apply a (biased) lognormal transformation of the density
    field by computing:

    .. math::

        F(\delta) = \frac{1}{N} e^{b*\delta}

    where :math:`\delta` is the initial overdensity field and the
    normalization :math:`N` is chosen such that
    :math:`\langle F(\delta) \rangle = 1`

    Parameters
    ----------
    density : array_like
        the input density field to apply the transformation to
    bias : float, optional
        optionally apply a linear bias to the density field;
        default is unbiased (1.0)

    Returns
    -------
    toret : RealField
        the real field holding the transformed density field
    """
    toret = density.copy()
    toret[:] = numpy.exp(bias * density.value)
    toret[:] /= toret.cmean(dtype='f8')

    return toret


def poisson_sample_to_points(delta, displacement, pm, nbar, bias=1., seed=None, logger=None):
    """
    Poisson sample the linear delta and displacement fields to points.

    The steps in this function:

    #.  Apply a biased, lognormal transformation to the input ``delta`` field
    #.  Poisson sample the overdensity field to discrete points
    #.  Disribute the positions of particles uniformly within the mesh cells,
        and assign the displacement field at each cell to the particles

    Parameters
    ----------
    delta : RealField
        the linear overdensity field to sample
    displacement : list of RealField (3,)
        the linear displacement fields which is used to move the particles
    nbar : float
        the desired number density of the output catalog of objects
    bias : float, optional
        apply a linear bias to the overdensity field (default is 1.)
    seed : int, optional
        the random seed used to Poisson sample the field to points

    Returns
    -------
    pos : array_like, (N, 3)
        the Cartesian positions of each of the generated particles
    displ : array_like, (N, 3)
        the displacement field sampled for each of the generated particles in the
        same units as the ``pos`` array
    """
    comm = delta.pm.comm

    # seed1 used for poisson sampling
    # seed2 used for uniform shift within a cell.
    seed1, seed2 = numpy.random.RandomState(seed).randint(0, 0xfffffff, size=2)

    # apply the lognormal transformation to the initial conditions density
    # this creates a positive-definite delta (necessary for Poisson sampling)
    lagrangian_bias = bias - 1.
    delta = lognormal_transform(delta, bias=lagrangian_bias)

    if logger and pm.comm.rank == 0:
        logger.info("Lognormal transformation done")

    # mean number of objects per cell
    H = delta.BoxSize / delta.Nmesh
    overallmean = H.prod() * nbar

    # number of objects in each cell (per rank, as a RealField)
    cellmean = delta * overallmean

    # create a random state with the input seed
    rng = MPIRandomState(seed=seed1, comm=comm, size=delta.size)

    # generate poissons. Note that we use ravel/unravel to
    # maintain MPI invariane.
    Nravel = rng.poisson(lam=cellmean.ravel())
    N = delta.pm.create(mode='real')
    N.unravel(Nravel)

    Ntot = N.csum()
    if logger and pm.comm.rank == 0:
        logger.info("Poisson sampling done, total number of objects is %d" % Ntot)

    pos_mesh = delta.pm.generate_uniform_particle_grid(shift=0.0)
    disp_mesh = numpy.empty_like(pos_mesh)

    # no need to do decompose because pos_mesh is strictly within the
    # local volume of the RealField.
    N_per_cell = N.readout(pos_mesh, resampler='nnb')
    for i in range(N.ndim):
        disp_mesh[:, i] = displacement[i].readout(pos_mesh, resampler='nnb')

    # fight round off errors, if any
    N_per_cell = numpy.int64(N_per_cell + 0.5)

    pos = pos_mesh.repeat(N_per_cell, axis=0)
    disp = disp_mesh.repeat(N_per_cell, axis=0)

    del pos_mesh
    del disp_mesh

    if logger and pm.comm.rank == 0:
        logger.info("catalog produced. Assigning in cell shift.")

    # generate linear ordering of the positions.
    # this should have been a method in pmesh, e.g. argument
    # to genereate_uniform_particle_grid(return_id=True);

    # FIXME: after pmesh update, remove this
    orderby = numpy.int64(pos[:, 0] / H[0] + 0.5)
    for i in range(1, delta.ndim):
        orderby[...] *= delta.Nmesh[i]
        orderby[...] += numpy.int64(pos[:, i] / H[i] + 0.5)

    # sort by ID to maintain MPI invariance.
    pos = mpsort.sort(pos, orderby=orderby, comm=comm)
    disp = mpsort.sort(disp, orderby=orderby, comm=comm)

    if logger and pm.comm.rank == 0:
        logger.info("sorting done")

    rng_shift = MPIRandomState(seed=seed2, comm=comm, size=len(pos))
    in_cell_shift = rng_shift.uniform(0, H[i], itemshape=(delta.ndim,))

    pos[...] += in_cell_shift
    pos[...] %= delta.BoxSize

    if logger and pm.comm.rank == 0:
        logger.info("catalog shifted.")

    return pos, disp



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from nbodykit.utils import ScatterArray, GatherArray, FrontPadArray
from numpy.testing import assert_array_equal
import os
import pytest

setup_logging("debug")

@MPITest([1])
def test_json_quantity(comm):

    from nbodykit.utils import JSONEncoder, JSONDecoder
    import json
    import tempfile

    cosmo = cosmology.Planck15

    # astropy quantity
    cosmo2 = cosmo.to_astropy()
    pos = numpy.ones(10, dtype=[('Position', ('f4',3))])
    data = {'m_nu':cosmo2.m_nu, 'H0':cosmo2.H0, 'a':numpy.float64(10.0), 'b':10, 'pos':pos}

    # write to file
    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'w') as ff:
        json.dump(data, ff, cls=JSONEncoder)

    # load from file
    with open(tmpfile, 'r') as ff:
        data2 = json.load(ff, cls=JSONDecoder)

    # test for equality
    for name in ['H0', 'm_nu']:
        assert data2[name].__class__ == data[name].__class__
        numpy.testing.assert_array_equal(data2[name], data[name])
    assert data['a'] == data2['a']
    assert data['b'] == data2['b']

    os.remove(tmpfile)

@MPITest([2])
def test_gather_array(comm):

    # object arrays must fail
    data1a = numpy.ones(10, dtype=[('test', 'f8')])
    data2a = numpy.ones(10, dtype='f8')

    data1 = GatherArray(data1a, comm, root=0)
    if comm.rank == 0:
        numpy.testing.assert_array_equal(data1['test'], 1)
        assert len(data1) == 10 * comm.size
    else:
        assert data1 is None

    data2 = GatherArray(data2a, comm, root=0)
    if comm.rank == 0:
        numpy.testing.assert_array_equal(data2, 1)
        assert len(data2) == 10 * comm.size
    else:
        assert data2 is None

    data2 = GatherArray(data2a, comm, root=Ellipsis)
    numpy.testing.assert_array_equal(data2, 1)
    assert len(data2) == 10 * comm.size

    data1 = GatherArray(data1a, comm, root=Ellipsis)
    numpy.testing.assert_array_equal(data1['test'], 1)
    assert len(data1) == 10 * comm.size


@MPITest([2])
def test_gather_objects(comm):

    # object arrays must fail
    data1 = numpy.ones(10, dtype=[('test', 'O')])
    data2 = numpy.ones(10, dtype='O')

    with pytest.raises(ValueError):
        data1 = GatherArray(data1, comm, root=0)

    with pytest.raises(ValueError):
        data2 = GatherArray(data2, comm, root=0)

    with pytest.raises(ValueError):
        data1 = GatherArray(data1, comm, root=Ellipsis)

    with pytest.raises(ValueError):
        data2 = GatherArray(data2, comm, root=Ellipsis)

@MPITest([2])
def test_scatter_objects(comm):

    # object arrays must fail
    if comm.rank == 0:
        data1 = numpy.ones(10, dtype=[('test', 'O')])
        data2 = numpy.ones(10, dtype='O')
    else:
        data1 = None
        data2 = None

    with pytest.raises(ValueError):
        data1 = ScatterArray(data1, comm, root=0)

    with pytest.raises(ValueError):
        data2 = ScatterArray(data2, comm, root=0)

@MPITest([2])
def test_gather_bad_data(comm):

    # data
    if comm.rank == 0:
        data = numpy.ones(10, dtype=[('a', 'f')])
    else:
        data = numpy.ones(10, dtype=[('b', 'f')])

    # fields mismatch
    with pytest.raises(ValueError):
        data = GatherArray(data, comm, root=0)

    with pytest.raises(ValueError):
        data = GatherArray(data, comm, root=Ellipsis)

@MPITest([2])
def test_gather_bad_dtype(comm):

    # data
    if comm.rank == 0:
        data = numpy.ones(10, dtype=[('a', 'f4')])
    else:
        data = numpy.ones(10, dtype=[('a', 'f8')])

    # shape mismatch
    with pytest.raises(ValueError):
        data = GatherArray(data, comm, root=0)

    with pytest.raises(ValueError):
        data = GatherArray(data, comm, root=Ellipsis)

@MPITest([2])
def test_gather_bad_shape(comm):

    # data
    if comm.rank == 0:
        data = numpy.ones((10,2))
    else:
        data = numpy.ones((10,3))

    # shape mismatch
    with pytest.raises(ValueError):
        data = GatherArray(data, comm, root=0)

    with pytest.raises(ValueError):
        data = GatherArray(data, comm, root=Ellipsis)

@MPITest([2])
def test_gather_list(comm):

    # data
    data = numpy.ones(10, dtype=[('a', 'f')])

    # can't gather a list
    with pytest.raises(ValueError):
        data = GatherArray(list(data), comm, root=0)

    with pytest.raises(ValueError):
        data = GatherArray(list(data), comm, root=Ellipsis)

@MPITest([2])
def test_scatter_list(comm):

    # data
    if comm.rank == 0:
        data = list(numpy.ones(10, dtype=[('a', 'f')]))
    else:
        data = None

    # can't scatter list
    with pytest.raises(ValueError):
        data = ScatterArray(data, comm, root=0)


@MPITest([2])
def test_scatter_wrong_counts(comm):

    # data
    if comm.rank == 0:
        data = numpy.ones(10, dtype=[('a', 'f')])
    else:
        data = None

    # wrong counts length
    with pytest.raises(ValueError):
        data = ScatterArray(data, comm, root=0, counts=[0, 5, 5])

    # wrong counts sum
    with pytest.raises(ValueError):
        data = ScatterArray(data, comm, root=0, counts=[5, 7])

@MPITest([4])
def test_frontpad_array(comm):

    # object arrays must fail
    data1 = numpy.ones(10) * comm.rank

    data2 = FrontPadArray(data1, comm.rank * 2, comm)
    assert len(data2) == len(data1) + comm.rank * 2
    assert_array_equal(data2[:comm.rank * 2], comm.rank - 1)
    assert_array_equal(data2[comm.rank * 2:], data1)

    data2 = FrontPadArray(data1, comm.rank * 10, comm)
    assert len(data2) == len(data1) + comm.rank * 10

    assert_array_equal(data2[comm.rank * 10:], data1)

    for i in range(comm.rank):
        assert_array_equal(data2[(comm.rank - i - 1) * 10:(comm.rank - i)* 10], comm.rank - i - 1)



import numpy
import logging
import time
import warnings

from nbodykit import CurrentMPIComm
from nbodykit.utils import timer
from nbodykit.binned_statistic import BinnedStatistic
from nbodykit.algorithms.fftpower import project_to_basis, _find_unique_edges
from pmesh.pm import ComplexField

def get_real_Ylm(l, m):
    """
    Return a function that computes the real spherical
    harmonic of order (l,m)

    Parameters
    ----------
    l : int
        the degree of the harmonic
    m : int
        the order of the harmonic; abs(m) <= l

    Returns
    -------
    Ylm : callable
        a function that takes 4 arguments: (xhat, yhat, zhat)
        unit-normalized Cartesian coordinates and returns the
        specified Ylm

    References
    ----------
    https://en.wikipedia.org/wiki/Spherical_harmonics#Real_form
    """
    import sympy as sp

    # make sure l,m are integers
    l = int(l); m = int(m)

    # the relevant cartesian and spherical symbols
    x, y, z, r = sp.symbols('x y z r', real=True, positive=True)
    xhat, yhat, zhat = sp.symbols('xhat yhat zhat', real=True, positive=True)
    phi, theta = sp.symbols('phi theta')
    defs = [(sp.sin(phi), y/sp.sqrt(x**2+y**2)),
            (sp.cos(phi), x/sp.sqrt(x**2+y**2)),
            (sp.cos(theta), z/sp.sqrt(x**2 + y**2+z**2))]

    # the normalization factors
    if m == 0:
        amp = sp.sqrt((2*l+1) / (4*numpy.pi))
    else:
        amp = sp.sqrt(2*(2*l+1) / (4*numpy.pi) * sp.factorial(l-abs(m)) / sp.factorial(l+abs(m)))

    # the cos(theta) dependence encoded by the associated Legendre poly
    expr = (-1)**m * sp.assoc_legendre(l, abs(m), sp.cos(theta))

    # the phi dependence
    if m < 0:
        expr *= sp.expand_trig(sp.sin(abs(m)*phi))
    elif m > 0:
        expr *= sp.expand_trig(sp.cos(m*phi))

    # simplify
    expr = sp.together(expr.subs(defs)).subs(x**2 + y**2 + z**2, r**2)
    expr = amp * expr.expand().subs([(x/r, xhat), (y/r, yhat), (z/r, zhat)])
    Ylm = sp.lambdify((xhat,yhat,zhat), expr, 'numexpr')

    # attach some meta-data
    Ylm.expr = expr
    Ylm.l    = l
    Ylm.m    = m

    return Ylm

class ConvolvedFFTPower(object):
    """
    Algorithm to compute power spectrum multipoles using FFTs
    for a data survey with non-trivial geometry.

    Due to the geometry, the estimator computes the true power spectrum
    convolved with the window function (FFT of the geometry).

    This estimator implemented in this class is described in detail in
    Hand et al. 2017 (arxiv:1704.02357). It uses the spherical harmonic
    addition theorem such that only :math:`2\ell+1` FFTs are required to
    compute each multipole. This differs from the implementation in
    Bianchi et al. and Scoccimarro et al., which requires
    :math:`(\ell+1)(\ell+2)/2` FFTs.

    Results are computed when the object is inititalized, and the result is
    stored in the :attr:`poles` attribute. Important meta-data computed
    during algorithm execution is stored in the :attr:`attrs` dict. See the
    documenation of :func:`~ConvolvedFFTPower.run`.

    .. note::
        A full tutorial on the class is available in the documentation
        :ref:`here <convpower>`.

    .. note::
        Cross correlations are only supported when the FKP weight column
        differs between the two mesh objects, i.e., the underlying ``data``
        and ``randoms`` must be the same. This allows users to compute
        the cross power spectrum of the same density field, weighted
        differently.

    Parameters
    ----------
    first : FKPCatalog, FKPCatalogMesh
        the first source to paint the data/randoms; FKPCatalog is automatically
        converted to a FKPCatalogMesh, using default painting parameters
    poles : list of int
        a list of integer multipole numbers ``ell`` to compute
    second : FKPCatalog, FKPCatalogMesh, optional
        the second source to paint the data/randoms; cross correlations are
        only supported when the weight column differs between the two mesh
        objects, i.e., the underlying ``data`` and ``randoms`` must be the same!
    kmin : float, optional
        the edge of the first wavenumber bin; default is 0
    dk : float, optional
        the spacing in wavenumber to use; if not provided; the fundamental mode
        of the box is used

    References
    ----------
    * Hand, Nick et al. `An optimal FFT-based anisotropic power spectrum estimator`, 2017
    * Bianchi, Davide et al., `Measuring line-of-sight-dependent Fourier-space clustering using FFTs`,
      MNRAS, 2015
    * Scoccimarro, Roman, `Fast estimators for redshift-space clustering`, Phys. Review D, 2015
    """
    logger = logging.getLogger('ConvolvedFFTPower')

    def __init__(self, first, poles,
                    second=None,
                    Nmesh=None,
                    kmin=0.,
                    dk=None,
                    use_fkp_weights=None,
                    P0_FKP=None):

        if use_fkp_weights is not None or P0_FKP is not None:
            raise ValueError("use_fkp_weights and P0_FKP are deprecated. Assign a FKPWeight column to source['randoms']['FKPWeight'] and source['data']['FKPWeight'] with the help of the FKPWeightFromNbar(nbar) function")

        first = _cast_mesh(first, Nmesh=Nmesh)
        if second is not None:
            second = _cast_mesh(second, Nmesh=Nmesh)
        else:
            second = first

        # data/randoms of second must be same as second
        # only difference can be FKP weight currently
        if not is_valid_crosscorr(first, second):
            msg = ("ConvolvedFFTPower cross-correlations currently require the same"
                   " FKPCatalog (data/randoms), such that only the weight column can vary")
            raise NotImplementedError(msg)

        self.first = first
        self.second = second

        # grab comm from first source
        self.comm = first.comm

        # check for comm mismatch
        assert second.comm is first.comm, "communicator mismatch between input sources"

        # make a box big enough for both catalogs if they are not equal
        # NOTE: both first/second must have the same BoxCenter to recenter Position
        if not numpy.array_equal(first.attrs['BoxSize'], second.attrs['BoxSize']):

            # stack box coordinates together
            joint = {}
            for name in ['BoxSize', 'BoxCenter']:
                joint[name] = numpy.vstack([first.attrs[name], second.attrs[name]])

            # determine max box length along each dimension
            argmax = numpy.argmax(joint['BoxSize'], axis=0)
            joint['BoxSize'] = joint['BoxSize'][argmax, [0,1,2]]
            joint['BoxCenter'] = joint['BoxCenter'][argmax, [0,1,2]]

            # re-center the box
            first.recenter_box(joint['BoxSize'], joint['BoxCenter'])
            second.recenter_box(joint['BoxSize'], joint['BoxCenter'])

        # make a list of multipole numbers
        if numpy.isscalar(poles):
            poles = [poles]

        # store meta-data
        self.attrs = {}
        self.attrs['poles'] = poles
        self.attrs['dk'] = dk
        self.attrs['kmin'] = kmin

        # store BoxSize and BoxCenter from source
        self.attrs['Nmesh'] = self.first.attrs['Nmesh'].copy()
        self.attrs['BoxSize'] = self.first.attrs['BoxSize']
        self.attrs['BoxPad'] = self.first.attrs['BoxPad']
        self.attrs['BoxCenter'] = self.first.attrs['BoxCenter']

        # grab some mesh attrs, too
        self.attrs['mesh.resampler'] = self.first.resampler
        self.attrs['mesh.interlaced'] = self.first.interlaced

        # and run
        self.run()

    def run(self):
        """
        Compute the power spectrum multipoles. This function does not return
        anything, but adds several attributes (see below).

        Attributes
        ----------
        edges : array_like
            the edges of the wavenumber bins
        poles : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object that behaves similar to a structured array, with
            fancy slicing and re-indexing; it holds the measured multipole
            results, as well as the number of modes (``modes``) and average
            wavenumbers values in each bin (``k``)
        attrs : dict
            dictionary holding input parameters and several important quantites
            computed during execution:

            #. data.N, randoms.N :
                the unweighted number of data and randoms objects
            #. data.W, randoms.W :
                the weighted number of data and randoms objects, using the
                column specified as the completeness weights
            #. alpha :
                the ratio of ``data.W`` to ``randoms.W``
            #. data.norm, randoms.norm :
                the normalization of the power spectrum, computed from either
                the "data" or "randoms" catalog (they should be similar).
                See equations 13 and 14 of arxiv:1312.4611.
            #. data.shotnoise, randoms.shotnoise :
                the shot noise values for the "data" and "random" catalogs;
                See equation 15 of arxiv:1312.4611.
            #. shotnoise :
                the total shot noise for the power spectrum, equal to
                ``data.shotnoise`` + ``randoms.shotnoise``; this should be subtracted from
                the monopole.
            #. BoxSize :
                the size of the Cartesian box used to grid the data and
                randoms objects on a Cartesian mesh.

            For further details on the meta-data, see
            :ref:`the documentation <fkp-meta-data>`.
        """
        pm = self.first.pm

        # setup the binning in k out to the minimum nyquist frequency
        dk = 2*numpy.pi/pm.BoxSize.min() if self.attrs['dk'] is None else self.attrs['dk']

        if dk > 0:
            kedges = numpy.arange(self.attrs['kmin'], numpy.pi*pm.Nmesh.min()/pm.BoxSize.max() + dk/2, dk)
            kcoords = None
        else:
            k = pm.create_coords('complex')
            kedges, kcoords = _find_unique_edges(k, 2 * numpy.pi / pm.BoxSize, pm.comm)
            if self.comm.rank == 0:
                self.logger.info('%d unique k values are found' % len(kcoords))

        # measure the binned 1D multipoles in Fourier space
        result = self._compute_multipoles(kedges)

        # set all the necessary results
        self.poles = BinnedStatistic(['k'], [kedges], result,
                            fields_to_sum=['modes'],
                            coords=[kcoords],
                            **self.attrs)
        self.edges = kedges

    def to_pkmu(self, mu_edges, max_ell):
        """
        Invert the measured multipoles :math:`P_\ell(k)` into power
        spectrum wedges, :math:`P(k,\mu)`.

        Parameters
        ----------
        mu_edges : array_like
            the edges of the :math:`\mu` bins
        max_ell : int
            the maximum multipole to use when computing the wedges;
            all even multipoles with :math:`ell` less than or equal
            to this number are included

        Returns
        -------
        pkmu : BinnedStatistic
            a data set holding the :math:`P(k,\mu)` wedges
        """
        from scipy.special import legendre
        from scipy.integrate import quad

        def compute_coefficient(ell, mumin, mumax):
            """
            Compute how much each multipole contributes to a given wedges.
            This returns:

            .. math::
                \frac{1}{\mu_{max} - \mu_{max}} \int_{\mu_{min}}^{\mu^{max}} \mathcal{L}_\ell(\mu)
            """
            norm = 1.0 / (mumax - mumin)
            return norm * quad(lambda mu: legendre(ell)(mu), mumin, mumax)[0]

        # make sure we have all the poles measured
        ells = list(range(0, max_ell+1, 2))
        if any('power_%d' %ell not in self.poles for ell in ells):
            raise ValueError("measurements for ells=%s required if max_ell=%d" %(ells, max_ell))

        # new data array
        dtype = numpy.dtype([('power', 'c8'), ('k', 'f8'), ('mu', 'f8')])
        data = numpy.zeros((self.poles.shape[0], len(mu_edges)-1), dtype=dtype)

        # loop over each wedge
        bounds = list(zip(mu_edges[:-1], mu_edges[1:]))
        for imu, mulims in enumerate(bounds):

            # add the contribution from each Pell
            for ell in ells:
                coeff = compute_coefficient(ell, *mulims)
                data['power'][:,imu] += coeff * self.poles['power_%d' %ell]

            data['k'][:,imu] = self.poles['k']
            data['mu'][:,imu] = numpy.ones(len(data))*0.5*(mulims[1]+mulims[0])

        dims = ['k', 'mu']
        edges = [self.poles.edges['k'], mu_edges]
        return BinnedStatistic(dims=dims, edges=edges, data=data, coords=[self.poles.coords['k'], None], **self.attrs)

    def __getstate__(self):
        state = dict(poles=self.poles.__getstate__(),
                     attrs=self.attrs)
        return state

    def __setstate__(self, state):
        self.attrs = state['attrs']
        self.poles = BinnedStatistic.from_state(state['poles'])

    def __setstate_pre000305__(self, state):
        """ compatible version of setstate for files generated before 0.3.5 """
        edges = state['edges']
        sefl.attrs = state['attrs']
        self.poles = BinnedStatistic(['k'], [edges], self.poles, fields_to_sum=['modes'])

    def save(self, output):
        """
        Save the ConvolvedFFTPower result to disk.

        The format is currently json.

        Parameters
        ----------
        output : str
            the name of the file to dump the JSON results to
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('saving ConvolvedFFTPower result to %s' %output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, output, comm=None, format='current'):
        """
        Load a saved ConvolvedFFTPower result, which has been saved to
        disk with :func:`ConvolvedFFTPower.save`.

        The current MPI communicator is automatically used
        if the ``comm`` keyword is ``None``

        format can be 'current', or 'pre000305' for files generated before 0.3.5.

        """
        import json
        from nbodykit.utils import JSONDecoder

        if comm.rank == 0:
            with open(output, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)

        if format == 'current':
            self.__setstate__(state)
        elif format == 'pre000305':
            self.__setstate_pre000305__(state)

        self.comm = comm
        return self

    def _compute_multipoles(self, kedges):
        """
        Compute the window-convoled power spectrum multipoles, for a data set
        with non-trivial survey geometry.

        This estimator builds upon the work presented in Bianchi et al. 2015
        and Scoccimarro et al. 2015, but differs in the implementation. This
        class uses the spherical harmonic addition theorem such that
        only :math:`2\ell+1` FFTs are required per multipole, rather than the
        :math:`(\ell+1)(\ell+2)/2` FFTs in the implementation presented by
        Bianchi et al. and Scoccimarro et al.

        References
        ----------
        * Bianchi, Davide et al., `Measuring line-of-sight-dependent Fourier-space clustering using FFTs`,
          MNRAS, 2015
        * Scoccimarro, Roman, `Fast estimators for redshift-space clustering`, Phys. Review D, 2015
        """
        # clear compensation from the actions
        for source in [self.first, self.second]:
            source.actions[:] = []; source.compensated = False
            assert len(source.actions) == 0

        # compute the compensations
        compensation = {}
        for name, mesh in zip(['first', 'second'], [self.first, self.second]):
            compensation[name] = get_compensation(mesh)
            if self.comm.rank == 0:
                if compensation[name] is not None:
                    args = (compensation[name]['func'].__name__, name)
                    self.logger.info("using compensation function %s for source '%s'" % args)
                else:
                    self.logger.warning("no compensation applied for source '%s'" % name)

        rank = self.comm.rank
        pm   = self.first.pm

        # setup the 1D-binning
        muedges = numpy.linspace(0, 1, 2, endpoint=True)
        edges = [kedges, muedges]

        # make a structured array to hold the results
        cols   = ['k'] + ['power_%d' %l for l in sorted(self.attrs['poles'])] + ['modes']
        dtype  = ['f8'] + ['c8']*len(self.attrs['poles']) + ['i8']
        dtype  = numpy.dtype(list(zip(cols, dtype)))
        result = numpy.empty(len(kedges)-1, dtype=dtype)

        # offset the box coordinate mesh ([-BoxSize/2, BoxSize]) back to
        # the original (x,y,z) coords
        offset = self.attrs['BoxCenter'] + 0.5*pm.BoxSize / pm.Nmesh

        # always need to compute ell=0
        poles = sorted(self.attrs['poles'])
        if 0 not in poles:
            poles = [0] + poles
        assert poles[0] == 0

        # spherical harmonic kernels (for ell > 0)
        Ylms = [[get_real_Ylm(l,m) for m in range(-l, l+1)] for l in poles[1:]]

        # paint the 1st FKP density field to the mesh (paints: data - alpha*randoms, essentially)
        rfield1 = self.first.compute(Nmesh=self.attrs['Nmesh'])
        meta1 = rfield1.attrs.copy()
        if rank == 0:
            self.logger.info("%s painting of 'first' done" %self.first.resampler)

        # store alpha: ratio of data to randoms
        self.attrs['alpha'] = meta1['alpha']

        # FFT 1st density field and apply the resampler transfer kernel
        cfield = rfield1.r2c()
        if compensation['first'] is not None:
            cfield.apply(out=Ellipsis, **compensation['first'])
        if rank == 0: self.logger.info('ell = 0 done; 1 r2c completed')

        # monopole A0 is just the FFT of the FKP density field
        # NOTE: this holds FFT of density field #1
        volume = pm.BoxSize.prod()
        A0_1 = ComplexField(pm)
        A0_1[:] = cfield[:] * volume # normalize with a factor of volume

        # paint second mesh too?
        if self.first is not self.second:

            # paint the second field
            rfield2 = self.second.compute(Nmesh=self.attrs['Nmesh'])
            meta2 = rfield2.attrs.copy()
            if rank == 0: self.logger.info("%s painting of 'second' done" %self.second.resampler)

            # need monopole of second field
            if 0 in self.attrs['poles']:

                # FFT density field and apply the resampler transfer kernel
                A0_2 = rfield2.r2c()
                A0_2[:] *= volume
                if compensation['second'] is not None:
                    A0_2.apply(out=Ellipsis, **compensation['second'])
        else:
            rfield2 = rfield1
            meta2 = meta1

            # monopole of second field is first field
            if 0 in self.attrs['poles']:
                A0_2 = A0_1

        # ensure alpha from first mesh is equal to alpha from second mesh
        # NOTE: this is mostly just a sanity check, and should always be true if
        # we made it this far already
        if not numpy.allclose(rfield1.attrs['alpha'], rfield2.attrs['alpha'], rtol=1e-3):
            msg = ("ConvolvedFFTPower cross-correlations currently require the same"
                   " FKPCatalog (data/randoms), such that only the weight column can vary;"
                   " different ``alpha`` values found for first/second meshes")
            raise ValueError(msg)

        # save the painted density field #2 for later
        density2 = rfield2.copy()

        # initialize the memory holding the Aell terms for
        # higher multipoles (this holds sum of m for fixed ell)
        # NOTE: this will hold FFTs of density field #2
        Aell = ComplexField(pm)

        # the real-space grid
        xgrid = [xx.astype('f8') + offset[ii] for ii, xx in enumerate(density2.slabs.optx)]
        xnorm = numpy.sqrt(sum(xx**2 for xx in xgrid))
        xgrid = [x/xnorm for x in xgrid]

        # the Fourier-space grid
        kgrid = [kk.astype('f8') for kk in cfield.slabs.optx]
        knorm = numpy.sqrt(sum(kk**2 for kk in kgrid)); knorm[knorm==0.] = numpy.inf
        kgrid = [k/knorm for k in kgrid]

        # proper normalization: same as equation 49 of Scoccimarro et al. 2015
        for name in ['data', 'randoms']:
            self.attrs[name+'.norm'] = self.normalization(name, self.attrs['alpha'])

        if self.attrs['randoms.norm'] > 0:
            norm = 1.0 / self.attrs['randoms.norm']

            # check normalization
            Adata = self.attrs['data.norm']
            Aran = self.attrs['randoms.norm']
            if not numpy.allclose(Adata, Aran, rtol=0.05):
                msg = "normalization in ConvolvedFFTPower different by more than 5%; "
                msg += ",algorithm requires they must be similar\n"
                msg += "\trandoms.norm = %.6f, data.norm = %.6f\n" % (Aran, Adata)
                msg += "\tpossible discrepancies could be related to normalization "
                msg += "of n(z) column ('%s')\n" % self.first.nbar
                msg += "\tor the consistency of the FKP weight column for 'data' "
                msg += "and 'randoms';\n"
                msg += "\tn(z) columns for 'data' and 'randoms' should be "
                msg += "normalized to represent n(z) of the data catalog"
                raise ValueError(msg)

            if rank == 0:
                self.logger.info("normalized power spectrum with `randoms.norm = %.6f`" % Aran)
        else:
            # an empty random catalog is provides, so we will ignore the normalization.
            norm = 1.0
            if rank == 0:
                self.logger.info("normalization of power spectrum is neglected, as no random is provided.")


        # loop over the higher order multipoles (ell > 0)
        start = time.time()
        for iell, ell in enumerate(poles[1:]):

            # clear 2D workspace
            Aell[:] = 0.

            # iterate from m=-l to m=l and apply Ylm
            substart = time.time()
            for Ylm in Ylms[iell]:

                # reset the real-space mesh to the original density #2
                rfield2[:] = density2[:]

                # apply the config-space Ylm
                for islab, slab in enumerate(rfield2.slabs):
                    slab[:] *= Ylm(xgrid[0][islab], xgrid[1][islab], xgrid[2][islab])

                # real to complex of field #2
                rfield2.r2c(out=cfield)

                # apply the Fourier-space Ylm
                for islab, slab in enumerate(cfield.slabs):
                    slab[:] *= Ylm(kgrid[0][islab], kgrid[1][islab], kgrid[2][islab])

                # add to the total sum
                Aell[:] += cfield[:]

                # and this contribution to the total sum
                substop = time.time()
                if rank == 0:
                    self.logger.debug("done term for Y(l=%d, m=%d) in %s" %(Ylm.l, Ylm.m, timer(substart, substop)))

            # apply the compensation transfer function
            if compensation['second'] is not None:
                Aell.apply(out=Ellipsis, **compensation['second'])

            # factor of 4*pi from spherical harmonic addition theorem + volume factor
            Aell[:] *= 4*numpy.pi*volume

            # log the total number of FFTs computed for each ell
            if rank == 0:
                args = (ell, len(Ylms[iell]))
                self.logger.info('ell = %d done; %s r2c completed' %args)

            # calculate the power spectrum multipoles, slab-by-slab to save memory
            # NOTE: this computes (A0 of field #1) * (Aell of field #2).conj()
            for islab in range(A0_1.shape[0]):
                Aell[islab,...] = norm * A0_1[islab] * Aell[islab].conj()

            # project on to 1d k-basis (averaging over mu=[0,1])
            proj_result, _ = project_to_basis(Aell, edges)
            result['power_%d' %ell][:] = numpy.squeeze(proj_result[2])

        # summarize how long it took
        stop = time.time()
        if rank == 0:
            self.logger.info("higher order multipoles computed in elapsed time %s" %timer(start, stop))

        # also compute ell=0
        if 0 in self.attrs['poles']:

            # the 3D monopole
            for islab in range(A0_1.shape[0]):
                A0_1[islab,...] = norm*A0_1[islab]*A0_2[islab].conj()

            # the 1D monopole
            proj_result, _ = project_to_basis(A0_1, edges)
            result['power_0'][:] = numpy.squeeze(proj_result[2])

        # save the number of modes and k
        result['k'][:] = numpy.squeeze(proj_result[0])
        result['modes'][:] = numpy.squeeze(proj_result[-1])

        # compute shot noise
        self.attrs['shotnoise'] = self.shotnoise(self.attrs['alpha'])

        # copy over any painting meta data
        if self.first is self.second:
            copy_meta(self.attrs, meta1)
        else:
            copy_meta(self.attrs, meta1, prefix='first')
            copy_meta(self.attrs, meta2, prefix='second')

        return result

    def normalization(self, name, alpha):
        r"""
        Compute the power spectrum normalization, using either the
        ``data`` or ``randoms`` source.

        The normalization is given by:

        .. math::

            A = \int d^3x \bar{n}'_1(x) \bar{n}'_2(x) w_{\mathrm{fkp},1} w_{\mathrm{fkp},2}.

        The mean densities are assumed to be the same, so this can be converted
        to a summation over objects in the source, as

        .. math::

            A = \sum w_{\mathrm{comp},1} \bar{n}_2 w_{\mathrm{fkp},1} w_{\mathrm{fkp},2}.

        References
        ----------
        see Eqs. 13,14 of Beutler et al. 2014, "The clustering of galaxies in the
        SDSS-III Baryon Oscillation Spectroscopic Survey: testing gravity with redshift
        space distortions using the power spectrum multipoles"
        """
        assert name in ['data', 'randoms']

        if name+'.norm' not in self.attrs:

            # the selection (same for first/second)
            sel = self.first.source.compute(self.first.source[name][self.first.selection])

            # selected first/second meshes for "name" (data or randoms)
            first = self.first.source[name][sel]
            second = self.second.source[name][sel]

            # these are assumed the same b/w first and second meshes
            comp_weight = first[self.first.comp_weight]
            nbar = second[self.second.nbar]

            # different weights allowed for first and second mesh
            fkp_weight1 = first[self.first.fkp_weight]
            if self.second is self.first:
                fkp_weight2 = fkp_weight1
            else:
                fkp_weight2 = second[self.second.fkp_weight]

            A  = nbar*comp_weight*fkp_weight1*fkp_weight2
            if name == 'randoms':
                A *= alpha
            A = first.compute(A.sum())
            self.attrs[name+'.norm'] = self.comm.allreduce(A)

        return self.attrs[name+'.norm']

    def shotnoise(self, alpha):
        r"""
        Compute the power spectrum shot noise, using either the
        ``data`` or ``randoms`` source.

        This computes:

        .. math::

            S = \sum (w_\mathrm{comp} w_\mathrm{fkp})^2

        References
        ----------
        see Eq. 15 of Beutler et al. 2014, "The clustering of galaxies in the
        SDSS-III Baryon Oscillation Spectroscopic Survey: testing gravity with redshift
        space distortions using the power spectrum multipoles"
        """
        if 'shotnoise' not in self.attrs:

            Pshot = 0
            for name in ['data', 'randoms']:

                # the selection (same for first/second)
                sel = self.first.source.compute(self.first.source[name][self.first.selection])

                # selected first/second meshes for "name" (data or randoms)
                first = self.first.source[name][sel]
                second = self.second.source[name][sel]

                # completeness weights (assumed same for first/second)
                comp_weight = first[self.first.comp_weight]

                # different weights allowed for first and second mesh
                fkp_weight1 = first[self.first.fkp_weight]
                if self.first is self.second:
                    fkp_weight2 = fkp_weight1
                else:
                    fkp_weight2 = second[self.second.fkp_weight]

                S = (comp_weight**2*fkp_weight1*fkp_weight2).sum()
                if name == 'randoms':
                    S *= alpha**2
                Pshot += S # add to total

        # reduce sum across all ranks
        Pshot = self.comm.allreduce(first.compute(Pshot))

        # divide by normalization from randoms
        return Pshot / self.attrs['randoms.norm']

def _cast_mesh(mesh, Nmesh):
    """
    Cast an object to a MeshSource. Nmesh is used only on FKPCatalog

    """
    from .catalog import FKPCatalog
    from .catalogmesh import FKPCatalogMesh

    if not isinstance(mesh, (FKPCatalogMesh, FKPCatalog)):
        raise TypeError("input sources should be a FKPCatalog or FKPCatalogMesh")

    if isinstance(mesh, FKPCatalog):
        # if input is CatalogSource, use defaults to make it into a mesh
        mesh = mesh.to_mesh(Nmesh=Nmesh, dtype='f8', compensated=False)

    if Nmesh is not None and any(mesh.attrs['Nmesh'] != Nmesh):
        raise ValueError(("Mismatched Nmesh between __init__ and mesh.attrs; "
                          "if trying to re-sample with a different mesh, specify "
                          "`Nmesh` as keyword of to_mesh()"))

    return mesh

def get_compensation(mesh):
    toret = None
    try:
        compensation = mesh._get_compensation()
        toret = {'func':compensation[0][1], 'kind':compensation[0][2]}
    except ValueError:
        pass
    return toret

def copy_meta(attrs, meta, prefix=""):
    if prefix:
        prefix += '.'
    for key in meta:
        if key.startswith('data.') or key.startswith('randoms.'):
            attrs[prefix+key] = meta[key]

def is_valid_crosscorr(first, second):

    if second.source is not first.source:
        return False

    same_cols = ['selection', 'comp_weight', 'nbar']
    if any(getattr(second, name) != getattr(first, name) for name in same_cols):
        return False

    return True



from distutils.core import setup
from distutils.util import convert_path
import os

def find_version(path):
    import re
    # path shall be a plain ascii text file.
    s = open(path, 'rt').read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              s, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Version not found")

def find_packages(base_path):
    base_path = convert_path(base_path)
    found = []
    for root, dirs, files in os.walk(base_path, followlinks=True):
        dirs[:] = [d for d in dirs if d[0] != '.' and d not in ('ez_setup', '__pycache__')]
        relpath = os.path.relpath(root, base_path)
        parent = relpath.replace(os.sep, '.').lstrip('.')
        if relpath != '.' and parent not in found:
            # foo.bar package but no foo package, skip
            continue
        for dir in dirs:
            if os.path.isfile(os.path.join(root, dir, '__init__.py')):
                package = '.'.join((parent, dir)) if parent else dir
                found.append(package)
    return found

def find_test_data(base_path, pkg_name):
    data = []
    for root, dirs, files in os.walk(base_path):
        for d in dirs:
            if d == 'data' and root.split('/')[-1] == 'tests':
                path = os.path.join(root, 'data', '*')
                data.append(os.path.relpath(path, pkg_name))
    return data

# the base dependencies
with open('requirements.txt', 'r') as fh:
    dependencies = [l.strip() for l in fh]

# extra dependencies
extras = {}
with open('requirements-extras.txt', 'r') as fh:
    extras['extras'] = [l.strip() for l in fh][1:]
    extras['full'] = extras['extras'] #

# package data
data = find_test_data('nbodykit', 'nbodykit')
data.append('style/*mplstyle') # also install mplstyle files

setup(name="nbodykit",
      version=find_version("nbodykit/version.py"),
      author="Yu Feng, Nick Hand, et al",
      maintainer="Yu Feng",
      maintainer_email="rainwoodman@gmail.com",
      description="Analysis kit for large-scale structure datasets, the massively parallel way",
      url="http://github.com/bccp/nbodykit",
      zip_safe=False,
      package_dir = {'nbodykit': 'nbodykit'},
      package_data = {'nbodykit': data},
      packages = find_packages('.'),
      license='GPL3',
      install_requires=dependencies,
      extras_require=extras
)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_array_equal, assert_allclose
import kdcount.correlate as correlate
import os
import pytest

setup_logging()

def get_spherical_volume(source):
    pos = gather_data(source, "Position")
    rad = ((pos - 0.5*source.attrs['BoxSize'])**2).sum(axis=-1)**0.5
    keep = rad < 0.5 * source.attrs['BoxSize'].min()
    ra = gather_data(source, "RA")[keep]
    dec = gather_data(source, "DEC")[keep]
    return numpy.vstack([ra, dec]).T

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_sim_data(seed, comm):
    s = UniformCatalog(nbar=1000, BoxSize=1.0, seed=seed, comm=comm)
    s['RA'], s['DEC'] = transform.CartesianToEquatorial(s['Position'], observer=0.5*s.attrs['BoxSize'])
    return s

def reference_sim_tpcf(pos1, theta_edges, randoms=None, pos2=None):
    """Refernce angular 2PCF using halotools"""
    from halotools.mock_observables import angular_tpcf

    estimator = 'Natural' if randoms is None else 'Landy-Szalay'
    do_auto = True if pos2 is None else False
    return angular_tpcf(pos1, theta_edges, sample2=pos2, randoms=randoms,
                            estimator=estimator, do_auto=do_auto)

@MPITest([1, 4])
def test_sim_nonperiodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)
    randoms = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    theta_edges = numpy.linspace(0.1, 10.0, 20)

    # compute 2PCF
    r = SimulationBox2PCF('angular', source, theta_edges, periodic=False, randoms1=randoms)

    # verify with reference
    cf = [-1.409027, 0.381019, -0.4193353, -0.1228877, -0.1578114, -0.1555974, 0.02935334, 0.04872244, 0.005932454,
        -0.3038669, 0.03336205, -0.07142007, -0.08075831, -0.07201585, -0.09670292, -0.1443989, -0.1554421, -0.2775648, 0.01500372]

    # verify with halotools // uses N*N instead of N*(N-1) normalization
    #ra1, dec1 = gather_data(source, "RA"),  gather_data(source, "DEC")
    #ra2, dec2 = gather_data(randoms, "RA"),  gather_data(randoms, "DEC")
    #cf = reference_sim_tpcf(numpy.vstack([ra1,dec1]).T, theta_edges, randoms=numpy.vstack([ra2,dec2]).T)
    assert_allclose(cf, r.corr['corr'], rtol=1e-5, atol=1e-3)

@MPITest([1, 4])
def test_sim_periodic_cross(comm):

    # uniform source of particles
    data1 = generate_sim_data(seed=42, comm=comm)
    data2 = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    theta_edges = numpy.linspace(0.1, 10.0, 20)

    # compute 2PCF
    r = SimulationBox2PCF('angular', data1, theta_edges, periodic=True, data2=data2)

    # verify with halotools
    sample1 = get_spherical_volume(data1)
    sample2 = get_spherical_volume(data2)
    cf = reference_sim_tpcf(sample1, theta_edges, pos2=sample2)
    assert_allclose(cf, r.corr['corr'])

@MPITest([1, 4])
def test_survey_auto(comm):

    # uniform source of particles
    data = generate_sim_data(seed=42, comm=comm)
    randoms = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    theta_edges = numpy.linspace(0.1, 10.0, 20)

    # compute 2PCF
    r = SurveyData2PCF('angular', data, randoms, theta_edges)

    # verify with reference
    cf = [-1.409027, 0.381019, -0.4193353, -0.1228877, -0.1578114, -0.1555974, 0.02935334, 0.04872244, 0.005932454,
        -0.3038669, 0.03336205, -0.07142007, -0.08075831, -0.07201585, -0.09670292, -0.1443989, -0.1554421, -0.2775648, 0.01500372]

    # verify with halotools // uses N*N instead of N*(N-1) normalization
    #ra1, dec1 = gather_data(data, "RA"),  gather_data(data, "DEC")
    #ra2, dec2 = gather_data(randoms, "RA"),  gather_data(randoms, "DEC")
    #cf = reference_sim_tpcf(numpy.vstack([ra1,dec1]).T, theta_edges, randoms=numpy.vstack([ra2,dec2]).T)
    assert_allclose(cf, r.corr['corr'], rtol=1e-5, atol=1e-3)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from numpy.testing import assert_array_equal, assert_allclose
import pytest

# debug logging
setup_logging("debug")

@MPITest([1])
def test_fftcorr_poles(comm):

    source = UniformCatalog(nbar=3e-3, BoxSize=512., seed=42, comm=comm)

    r = FFTCorr(source, mode='2d', BoxSize=1024, Nmesh=32, poles=[0,2,4])
    pkmu = r.corr['corr'].real
    mono = r.poles['corr_0'].real

    modes_1d = r.corr['modes'].sum(axis=-1)
    mono_from_pkmu = numpy.nansum(pkmu*r.corr['modes'], axis=-1) / modes_1d

    assert_array_equal(modes_1d, r.poles['modes'])
    assert_allclose(mono_from_pkmu, mono)

@MPITest([1])
def test_fftcorr_unique(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    r = FFTCorr(source, mode='1d', Nmesh=32, dr=0)
    p = r.corr
    assert_allclose(p.coords['r'], p['r'], rtol=1e-6)

@MPITest([1])
def test_fftcorr_padding(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    r = FFTCorr(source, mode='1d', BoxSize=1024, Nmesh=32)
    assert r.attrs['N1'] != 0
    assert r.attrs['N2'] != 0

@MPITest([1])
def test_fftcorr_save(comm):

    source = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)

    r = FFTCorr(source, mode='2d', Nmesh=32)
    r.save('fftcorr-test.json')

    r2 = FFTCorr.load('fftcorr-test.json', comm=comm)

    assert_array_equal(r.corr['r'], r2.corr['r'])
    assert_array_equal(r.corr['corr'], r2.corr['corr'])
    assert_array_equal(r.corr['mu'], r2.corr['mu'])
    assert_array_equal(r.corr['modes'], r2.corr['modes'])


@MPITest([1])
def test_fftcorr_mismatch_boxsize(comm):

    cosmo = cosmology.Planck15

    # input sources
    source1 = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-4, BoxSize=1024., seed=42, comm=comm)

    r = FFTCorr(source1, second=source2, mode='1d', BoxSize=1024, Nmesh=32)

@MPITest([1])
def test_fftcorr_mismatch_boxsize_fail(comm):

    cosmo = cosmology.Planck15

    # input sources
    mesh1 = UniformCatalog(nbar=3e-4, BoxSize=512., seed=42, comm=comm).to_mesh(Nmesh=32)
    mesh2 = UniformCatalog(nbar=3e-4, BoxSize=1024., seed=42, comm=comm).to_mesh(Nmesh=32)

    # raises an exception b/c meshes have different box sizes
    with pytest.raises(ValueError):
        r = FFTCorr(mesh1, second=mesh2, mode='1d', BoxSize=1024, Nmesh=32)




from runtests.mpi import MPITest
from nbodykit.tutorials import download_example_data
from nbodykit import setup_logging, CurrentMPIComm
from six.moves.urllib.error import HTTPError
import pytest
import shutil
import os
import tempfile

setup_logging()

@pytest.mark.xfail(raises=HTTPError)
@MPITest([1])
def test_download_directory(comm):

    # download Gadget1P snapshots directory
    filename = 'Gadget1P'
    download_example_data(filename)

    # check that it worked
    assert filename in os.listdir('.')
    assert len(os.listdir(filename)) > 0

    # remove the downloaded file
    shutil.rmtree(filename)

@pytest.mark.xfail(raises=HTTPError)
@MPITest([1])
def test_download_file(comm):

    # download TPM snapshot file
    filename = 'tpm_1.0000.bin.00'
    download_example_data(filename)

    # check that it worked
    assert filename in os.listdir('.')

    # remove filename
    os.remove(filename)

@pytest.mark.xfail(raises=HTTPError)
@MPITest([1])
def test_download_failure(comm):

    filename = 'MISSING'
    with pytest.raises(ValueError):
        download_example_data(filename)

@pytest.mark.xfail(raises=HTTPError)
@MPITest([1])
def test_missing_dirname(comm):

    with pytest.raises(ValueError):
        download_example_data('Gadget1P', download_dirname='MISSING')

@pytest.mark.xfail(raises=HTTPError)
@MPITest([1])
def test_download_to_location(comm):

    # download Gadget1P snapshots directory to specific directory
    filename = 'Gadget1P'
    loc = tempfile.mkdtemp()
    download_example_data(filename, download_dirname=loc)

    # check that it worked
    assert filename in os.listdir(loc)

    # delete temp directory
    shutil.rmtree(loc)
